#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import math
from collections import deque
from enum import Enum
from typing import List, Optional

import numpy as np
import pandas as pd


class AverageType(Enum):
    SMA = "sma"
    EMA = "ema"

    @staticmethod
    def from_str(name: str):
        name_upper = name.upper()
        names = [e.name for e in AverageType]
        assert name_upper in names, "Unknown average type:" + name
        return AverageType[name_upper]


class QuantilesTracker:
    """
    Tracks the mean, standard deviation, and quantiles of a random variable

    Note:
        We store the samples in memory so be careful with large number of
        samples
    """

    def __init__(self):
        self._samples: List = []

    def update(self, val: float) -> None:
        self._samples.append(val)

    def quantile(self, p) -> float:
        if len(self._samples) == 0:
            return float("Inf")
        return np.quantile(self._samples, p)

    @property
    def median_val(self) -> float:
        return self.quantile(0.5)

    @property
    def lower_quartile_val(self) -> float:
        return self.quantile(0.25)

    @property
    def upper_quartile_val(self) -> float:
        return self.quantile(0.75)


class ModelSequenceNumberTracker:
    r"""
    Class to keep track of the current global sequence number and statistics.

    Keeps track of "model_seqnum," which indicates the checkpoint version of global model.

    Example:
        For example, if global model_seqnum is 10 and the local model_seqnum
        (i.e. model_seqnum of a particular device state) is 7, then we record
        the diff of model_seqnum, which is 3 in this case.
    """

    def __init__(self):
        self._current_model_seqnum = 0
        self.seqnum_diff_stats = RandomVariableStatsTracker()

    def increment(self) -> int:
        r"""
        Increments the global model_seqnum
        """
        self._current_model_seqnum += 1
        return self._current_model_seqnum

    def get_staleness_and_update_stats(self, client_seqnum: int) -> int:
        r"""
        Compares the current global model_seqnum with model_seqnum of
        a particular client
        """
        # Seqnum_diff will be 0 for sequential training.
        seqnum_diff = self._current_model_seqnum - client_seqnum
        self.seqnum_diff_stats.update(seqnum_diff)
        return seqnum_diff

    def print_stats(self) -> None:
        print(f"ModelSeqNum: {self.current_seqnum}")
        print(f"\tSeqnumDiff, {self.seqnum_diff_stats.as_str()}")

    def mean(self) -> float:
        r"""
        Returns the mean difference between the global seq_num and local seq_num
        """
        return self.seqnum_diff_stats.mean()

    def standard_deviation(self) -> float:
        r"""
        Returns the SD difference between the global seq_num and local seq_num
        """
        return self.seqnum_diff_stats.standard_deviation()

    @property
    def current_seqnum(self) -> int:
        r"""
        Current global model seq num
        """
        return self._current_model_seqnum


class RandomVariableStatsTracker:
    """Keeps track of mean, variance, min and max values of a random variable"""

    def __init__(self, tracks_quantiles: bool = False):
        self._sum: float = 0
        self._sum_squares: float = 0
        self._min_val: float = float("Inf")
        self._max_val: float = -float("Inf")
        self._num_samples: int = 0
        self._quant_tracker: Optional[QuantilesTracker] = (
            QuantilesTracker() if tracks_quantiles else None
        )

    def update(self, val: float) -> None:
        self._sum += val
        self._sum_squares += val * val
        self._min_val = min(self._min_val, val)
        self._max_val = max(self._max_val, val)
        self._num_samples += 1
        if self._quant_tracker is not None:
            self._quant_tracker.update(val)

    def mean(self) -> float:
        if not self._num_samples:
            return float("Inf")
        return self._sum / self._num_samples

    def standard_deviation(self) -> float:
        if not self._num_samples:
            return float("Inf")
        mean_sum_squares = self._sum_squares / self._num_samples
        mean_squared = self.mean() * self.mean()
        variance = mean_sum_squares - mean_squared
        return math.sqrt(variance + 1e-6)

    @property
    def num_samples(self) -> int:
        return self._num_samples

    @property
    def min_val(self) -> float:
        return self._min_val

    @property
    def max_val(self) -> float:
        return self._max_val

    @property
    def mean_val(self) -> float:
        return self.mean()

    @property
    def standard_deviation_val(self) -> float:
        return self.standard_deviation()

    @property
    def median_val(self) -> float:
        if self._quant_tracker is None:
            return float("inf")
        return self._quant_tracker.median_val

    @property
    def lower_quartile_val(self) -> float:
        if self._quant_tracker is None:
            return float("inf")
        return self._quant_tracker.lower_quartile_val

    @property
    def upper_quartile_val(self) -> float:
        if self._quant_tracker is None:
            return float("inf")
        return self._quant_tracker.upper_quartile_val

    def as_str(self) -> str:
        return (
            f"Mean:{self.mean():.3f}, "
            f"SD:{self.standard_deviation():.3f}, "
            f"Min:{self.min_val:.3f}, Max:{self.max_val:.3f}"
        )


class RandomVariableStatsTrackerMA(RandomVariableStatsTracker):
    """
    Tracks the simple moving or exponential moving mean and
    standard deviation of a random variable

    Note:
        We store the window_size number of samples in memory so please
        keep window_size to be a reasonable number
    """

    def __init__(self, window_size: int, mode=AverageType.SMA, decay_factor=0.5):
        super().__init__()
        self._samples = deque(maxlen=window_size)
        self.window_size = window_size
        self.decay_factor = decay_factor
        self.mode = mode

    def update(self, val: float) -> None:
        super().update(val)
        self._samples.append(val)

    def mean(self) -> float:
        if not self._num_samples:
            raise ValueError("There are no samples in tracker.")
        return (
            np.mean(self._samples)
            if self.mode == AverageType.SMA
            else pd.Series(self._samples).ewm(alpha=self.decay_factor).mean().iloc[-1]
        )

    def standard_deviation(self) -> float:
        if not self._num_samples:
            raise ValueError("There are no samples in tracker.")
        return (
            np.std(self._samples)
            if self.mode == AverageType.SMA
            else pd.Series(self._samples).ewm(alpha=self.decay_factor).std().iloc[-1]
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from enum import Enum

from flsim.privacy.privacy_engine import (
    TreePrivacyEngine,
    GaussianPrivacyEngine,
    PrivacySetting,
    IPrivacyEngine,
)


class NoiseType(Enum):
    TREE_NOISE = "tree"
    GAUSSIAN = "guassian"


class PrivacyEngineFactory:
    @classmethod
    def create(
        cls,
        privacy_setting: PrivacySetting,
        users_per_round: int,
        num_total_users: int,
        noise_type: NoiseType,
    ) -> IPrivacyEngine:
        if noise_type == NoiseType.TREE_NOISE:
            return TreePrivacyEngine(
                privacy_setting=privacy_setting,
                users_per_round=users_per_round,
                num_total_users=num_total_users,
            )
        else:
            return GaussianPrivacyEngine(
                privacy_setting=privacy_setting,
                users_per_round=users_per_round,
                num_total_users=num_total_users,
            )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc
import copy
from typing import Any, Dict, List, Optional, Tuple

from flsim.common.timeline import Timeline
from flsim.interfaces.batch_metrics import IFLBatchMetrics
from flsim.interfaces.metrics_reporter import (
    Channel,
    IFLMetricsReporter,
    Metric,
    TrainingStage,
)
from torch.utils.tensorboard import SummaryWriter


class FLMetricsReporter(IFLMetricsReporter, abc.ABC):
    """Note that Tensorboard will not work with FBLearner
    if you use this metrics reporter. This is the
    MetricsRepoorter without any FB-specific dependency.
    If you want to use Tensorboard within FBLearner or
    inside Bento, please take a look at FBFLMetricsReporter.
    """

    def __init__(self, channels: List[Channel], log_dir: Optional[str] = None):
        self.channels = channels
        self.log_dir = log_dir
        if Channel.TENSORBOARD in channels:
            self.set_summary_writer(log_dir)
        if Channel.STDOUT in channels:
            self.print = print
        self.losses = []
        self.num_examples_list = []
        self.predictions_list = []
        self.targets_list = []
        self.model_inputs_list = []
        self.latest_scores: Dict[str, Any] = {}
        self.best_eval_metrics = None

    def set_summary_writer(self, log_dir: Optional[str]):
        self.writer = SummaryWriter(log_dir=log_dir)

    def add_batch_metrics(self, metrics: IFLBatchMetrics) -> None:
        self.losses.append(metrics.loss.item())
        self.num_examples_list.append(metrics.num_examples)
        self.predictions_list.append(metrics.predictions)
        self.targets_list.append(metrics.targets)
        self.model_inputs_list.append(metrics.model_inputs)

    def aggregate(self, one_user_metrics):
        pass

    def report_metrics(
        self,
        reset: bool,
        stage: TrainingStage,
        extra_metrics: Optional[List[Metric]] = None,
        **kwargs,
    ) -> Tuple[Any, bool]:
        metrics = self._report_metrics(
            reset=reset, stage=stage, extra_metrics=extra_metrics, **kwargs
        )
        if stage != TrainingStage.EVAL:
            return (metrics, False)

        if self.best_eval_metrics is None or self.compare_metrics(
            metrics, self.best_eval_metrics
        ):
            self.best_eval_metrics = copy.deepcopy(metrics)
            return (metrics, True)
        else:
            return (metrics, False)

    def _report_metrics(
        self,
        reset: bool,
        stage: TrainingStage,
        extra_metrics: Optional[List[Metric]] = None,
        **kwargs,
    ) -> Any:
        timeline: Timeline = kwargs.get("timeline", Timeline(global_round=1))
        # handle legacy case when epoch was provided
        epoch = kwargs.get("epoch", 0)
        if epoch > 0 and timeline.global_round == 1:
            timeline = Timeline(epoch=epoch, round=1)
        eval_metrics = None

        training_stage_in_str = TrainingStage(stage).name.title()
        if len(self.losses) > 0:
            mean_loss = sum(self.losses) / len(self.losses)

            if Channel.STDOUT in self.channels:
                self.print(f"{timeline}, Loss/{training_stage_in_str}: {mean_loss}")
            if Channel.TENSORBOARD in self.channels:
                self.writer.add_scalar(
                    f"Loss/{training_stage_in_str}",
                    mean_loss,
                    timeline.global_round_num(),
                )

            scores = self.compute_scores()
            self.latest_scores = scores

            for score_name, score in scores.items():
                if Channel.STDOUT in self.channels:
                    self.print(
                        f"{timeline}, {score_name}/{training_stage_in_str}: {score}"
                    )
                if Channel.TENSORBOARD in self.channels:
                    self.writer.add_scalar(
                        f"{score_name}/{training_stage_in_str}",
                        score,
                        timeline.global_round_num(),
                    )
            eval_metrics = self.create_eval_metrics(
                scores, mean_loss, timeline=timeline, stage=stage
            )
        # handle misc reporting values
        metrics = extra_metrics or []
        for metric in metrics:
            value = Metric.to_dict(metric.value) if metric.is_compund else metric.value
            if Channel.STDOUT in self.channels:
                self.print(
                    f"{timeline}, {metric.name}/{training_stage_in_str}: {value}"
                )
            if Channel.TENSORBOARD in self.channels:
                self.writer.add_scalars(
                    f"{metric.name}/{training_stage_in_str}",
                    value,
                    timeline.global_round_num(),
                ) if metric.is_compund else self.writer.add_scalar(
                    f"{metric.name}/{training_stage_in_str}",
                    value,
                    timeline.global_round_num(),
                )

        if reset:
            self.reset()

        return eval_metrics

    def reset(self):
        self.losses = []
        self.num_examples_list = []
        self.predictions_list = []
        self.targets_list = []
        self.model_inputs_list = []

    def get_latest_scores(self) -> Dict[str, Any]:
        return self.latest_scores

    @abc.abstractmethod
    def compare_metrics(self, eval_metrics, best_metrics) -> bool:
        """One should provide concrete implementation of how to compare
        eval_metrics and best_metrics.
        Return True if eval_metrics is better than best_metrics
        """
        pass

    @abc.abstractmethod
    def compute_scores(self) -> Dict[str, Any]:
        """One should override this method to specify how to compute scores
        (e.g. accuracy) of the model based on metrics.
        Return dictionary where key is name of the scores and value is
        score.
        """
        pass

    @abc.abstractmethod
    def create_eval_metrics(
        self, scores: Dict[str, Any], total_loss: float, **kwargs
    ) -> Any:
        """One should provide concrete implementation of how to construct
        object that represents evaluation metrics based on scores and total
        loss. Most of the case, one would just pick one of the scores or
        total loss as the evaluation metrics to pick the better model, but
        this interface also allows s/he to make evaluation metrics more
        complex and use it in conjunction with compare_metrics() function
        to determine which metrics is the better one.
        """
        pass



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.


from typing import NamedTuple


EPS = 1e-10


class Timeline(NamedTuple):
    """
    A point in time of the simulation.

    This contains epoch, round, and rounds_per_epoch or
    global_round and rounds_per_epoch. All of these values
    start from 1 by flsim convention. If any of these
    need not be set the default value of 0 is assinged, which
    helps us handle collision cases / initialization errors.

    Timeline usually refers to how many epochs/rounds have *finished*
    """

    epoch: int = 0
    round: int = 0
    global_round: int = 0
    rounds_per_epoch: int = 1

    def global_round_num(self) -> int:
        r"""
        Returns the global round number starting from 1.

        For example, if round is 2 and epoch is 10 and num_rounds_per_epoch is 10
        it will return 102.
        """
        assert (self.epoch and self.round) or self.global_round
        return (
            self.global_round or (self.epoch - 1) * self.rounds_per_epoch + self.round
        )

    def as_float(self, offset: int = 0) -> float:
        r"""
        prints the time-line as a floating number.

        E.g. if round is 2 and epoch is 10 and num_rounds_per_epoch is 10
        it will return 10.2. By default uses the round value of the object
        but can also show a value for a run that is offset away from the
        current round.
        """
        return (self.global_round_num() + offset) / self.rounds_per_epoch

    def tick(self, tick_interval: float) -> bool:
        r"""
        Returns true 'tick_interval' times every epoch
        E.g: if tick_interval is 10, it returns true 10 times every epoch
        This function is useful for deciding when to report train/eval results
        """
        return (self.as_float() + EPS) // tick_interval > (
            self.as_float(-1) + EPS
        ) // tick_interval

    def __str__(self):
        assert (self.epoch > 0 and self.round > 0) or self.global_round > 0
        e = self.epoch or ((self.global_round - 1) // self.rounds_per_epoch + 1)
        r = self.round or (((self.global_round - 1) % self.rounds_per_epoch) + 1)
        gr = self.global_round or ((e - 1) * self.rounds_per_epoch + r)
        return f"(epoch = {e}, round = {r}, global round = {gr})"



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from io import BytesIO

import numpy as np
import torch
import torch.utils.data as data


class DummyImageDataset(data.Dataset):
    def __init__(
        self,
        num_classes=10,
        num_images_per_class=10,
        num_channels=1,
        image_dim=(28, 28),
    ):
        self.num_classes = num_classes
        self.num_images_per_class = num_images_per_class
        if num_channels == 1:
            data_dim = (
                self.num_classes * self.num_images_per_class,
                image_dim[0],
                image_dim[1],
            )
        else:
            data_dim = (
                self.num_classes * self.num_images_per_class,
                num_channels,
                image_dim[0],
                image_dim[1],
            )

        self.data = torch.from_numpy(
            np.random.uniform(low=0.0, high=1.0, size=data_dim)
        )
        self.targets = list(range(self.num_classes))
        self.labels = torch.LongTensor(self.targets * self.num_images_per_class)

    def get_dataset(self):
        stream = BytesIO()
        torch.save((self.data, self.labels), stream)
        return stream

    def __len__(self):
        return self.num_classes * self.num_images_per_class

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx].item()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Any, Iterable, Optional, Type

import torch
from flsim.data.data_sharder import FLDataSharder
from flsim.interfaces.data_loader import IFLDataLoader
from flsim.interfaces.dataset import FLDataset
from torch.utils.data import DataLoader, Dataset


class FLDatasetDataLoader(IFLDataLoader):
    def __init__(
        self,
        dataset_class: Type[FLDataset],
        train_path: str,
        test_path: str,
        eval_path: str,
        sharder: FLDataSharder,
    ):
        # pyre-fixme[19]: Expected 0 positional arguments.
        self.train_dataset = dataset_class(train_path)
        # pyre-fixme[19]: Expected 0 positional arguments.
        self.test_dataset = dataset_class(test_path)
        # pyre-fixme[19]: Expected 0 positional arguments.
        self.eval_dataset = dataset_class(eval_path)
        self.sharder = sharder

    def fl_train_set(self, **kwargs) -> Iterable[Iterable[Any]]:
        train_batches = []
        for sharded_rows in self.sharder.shard_rows(self.train_dataset):
            train_batches.append(sharded_rows)
        return train_batches

    def fl_eval_set(self, **kwargs) -> Iterable[Any]:
        return self.eval_dataset

    def fl_test_set(self, **kwargs) -> Iterable[Any]:
        return self.test_dataset


class FLDatasetDataLoaderWithBatch(IFLDataLoader):
    # TODO: (jesikmin) T58865105 [papaya][data] Add shuffling support to FL
    # dataset data loader
    def __init__(
        self,
        train_dataset: Dataset,
        test_dataset: Dataset,
        eval_dataset: Dataset,
        sharder: FLDataSharder,
        train_batch_size: int,
        eval_batch_size: Optional[int] = None,
        test_batch_size: Optional[int] = None,
    ):
        assert train_batch_size > 0, "Batch size should be a positive integer."
        self.train_dataset = train_dataset
        self.test_dataset = test_dataset
        self.eval_dataset = eval_dataset
        self.sharder = sharder
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size
        self.test_batch_size = test_batch_size
        self._num_total_users: int = -1

    @property
    def num_total_users(self):
        assert (
            self._num_total_users != -1
        ), "num_total_users is valid only after fl_train_set() has been called"
        return self._num_total_users

    def fl_train_set(self, **kwargs) -> Iterable[Iterable[Any]]:
        self._num_total_users = 0
        rank = kwargs.get("rank", 0)
        world_size = kwargs.get("world_size", 1)

        train_batches = [
            user_data for _, user_data in self.sharder.shard_rows(self.train_dataset)
        ]
        # batch train_batches collected above
        final_train_batches = []
        # fetch attributes for each row
        keys = list(train_batches[0][0].keys())
        for one_user_data in train_batches:
            batched_user_data = []
            for i, single_data in enumerate(one_user_data):
                if i % self.train_batch_size == 0:
                    batched_user_data.append([])
                batched_user_data[-1].append(single_data)

            new_batched_user_data = []
            for a_batched_user_data in batched_user_data:
                batched_data_rows = {}
                for key in keys:
                    batched_data_rows[key] = []
                for single_user_data in a_batched_user_data:
                    for key in keys:
                        batched_data_rows[key].append(single_user_data[key])

                for key in keys:
                    batched_data_rows[key] = torch.stack(batched_data_rows[key])

                new_batched_user_data.append(batched_data_rows)
            # divide the total number of users evenly into world_size # of workers
            if self.num_total_users % world_size == rank:
                final_train_batches.append(new_batched_user_data)
            # count the total number of users
            self._num_total_users += 1

        return final_train_batches

    def fl_eval_set(self, **kwargs) -> Iterable[Any]:
        collate_fn = kwargs.get("collate_fn", None)  # identity function
        return DataLoader(
            self.eval_dataset, batch_size=self.eval_batch_size, collate_fn=collate_fn
        )

    def fl_test_set(self, **kwargs) -> Iterable[Any]:
        collate_fn = kwargs.get("collate_fn", None)  # identity function
        return DataLoader(
            self.test_dataset, batch_size=self.test_batch_size, collate_fn=collate_fn
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
from typing import Any, Optional, Tuple

import numpy as np
import torch
from flsim.clients.base_client import ClientConfig
from flsim.common.timeout_simulator import (
    NeverTimeOutSimulatorConfig,
    TimeOutSimulatorConfig,
)
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.metrics_reporter import IFLMetricsReporter
from flsim.interfaces.model import IFLModel
from flsim.optimizers.async_aggregators import AsyncAggregatorConfig
from flsim.optimizers.local_optimizers import LocalOptimizerSGDConfig
from flsim.optimizers.optimizer_scheduler import (
    OptimizerSchedulerConfig,
    ConstantLRSchedulerConfig,
)
from flsim.tests.utils import (
    FakeMetricReporter,
    verify_models_equivalent_after_training,
)
from flsim.trainers.async_trainer import AsyncTrainer, AsyncTrainerConfig
from flsim.utils.async_trainer.async_example_weights import (
    EqualExampleWeightConfig,
    AsyncExampleWeightConfig,
)
from flsim.utils.async_trainer.async_staleness_weights import (
    ConstantStalenessWeightConfig,
    AsyncStalenessWeightConfig,
)
from flsim.utils.async_trainer.async_user_selector import AsyncUserSelectorType
from flsim.utils.async_trainer.async_weights import AsyncWeightConfig
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGeneratorConfig,
    ConstantAsyncTrainingStartTimeDistrConfig,
    EventGeneratorConfig,
)
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from flsim.utils.tests.helpers.test_utils import FLTestUtils
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
)
from omegaconf import OmegaConf

CONSTANT_LR_SCHEDULER_CONFIG = ConstantLRSchedulerConfig()
EQUAL_EXAMPLE_WEIGHT_CONFIG = EqualExampleWeightConfig()
CONSTANT_STALENESS_WEIGHT_CONFIG = ConstantStalenessWeightConfig()


def create_event_generator_config(
    training_rate: int, training_duration_mean: float, training_duration_sd: float
) -> EventGeneratorConfig:
    return AsyncTrainingEventGeneratorConfig(
        training_start_time_distribution=ConstantAsyncTrainingStartTimeDistrConfig(
            training_rate=training_rate
        ),
        duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=training_duration_mean,
            training_duration_sd=training_duration_sd,
        ),
    )


def create_async_trainer(
    model: IFLModel,
    local_lr: float,
    epochs: int,
    event_generator_config: EventGeneratorConfig,
    aggregator_config: AsyncAggregatorConfig,
    example_weight_config: AsyncExampleWeightConfig = EQUAL_EXAMPLE_WEIGHT_CONFIG,
    local_lr_scheduler_config: OptimizerSchedulerConfig = CONSTANT_LR_SCHEDULER_CONFIG,
    staleness_weight_config: AsyncStalenessWeightConfig = CONSTANT_STALENESS_WEIGHT_CONFIG,
    timeout_simulator_config: Optional[TimeOutSimulatorConfig] = None,
    max_staleness: float = 1e10,
    report_train_metrics_after_aggregation: bool = False,
    eval_epoch_frequency: float = 1.0,
    always_keep_trained_model: bool = False,
    do_eval: bool = False,
    train_metrics_reported_per_epoch: int = 1,
):
    async_trainer = AsyncTrainer(
        model=model,
        cuda_enabled=False,
        **OmegaConf.structured(
            AsyncTrainerConfig(
                aggregator=aggregator_config,
                client=ClientConfig(
                    epochs=1,
                    optimizer=LocalOptimizerSGDConfig(
                        lr=local_lr,
                    ),
                    lr_scheduler=local_lr_scheduler_config,
                ),
                epochs=epochs,
                training_event_generator=event_generator_config,
                async_user_selector_type=AsyncUserSelectorType.ROUND_ROBIN,
                async_weight=AsyncWeightConfig(
                    example_weight=example_weight_config,
                    staleness_weight=staleness_weight_config,
                ),
                timeout_simulator=timeout_simulator_config
                or NeverTimeOutSimulatorConfig(),
                max_staleness=max_staleness,
                do_eval=do_eval,
                report_train_metrics_after_aggregation=report_train_metrics_after_aggregation,
                eval_epoch_frequency=eval_epoch_frequency,
                always_keep_trained_model=always_keep_trained_model,
                train_metrics_reported_per_epoch=train_metrics_reported_per_epoch,
            )
        ),
    )
    return async_trainer


def get_nonfl_optimizer(
    nonfl_model: IFLModel,
    fl_local_lr: float,
    fl_aggregator_config: AsyncAggregatorConfig,
) -> torch.optim.Optimizer:
    """Given FL trainer settings (local_lr, aggregator config),
    return a non-fl Optimizer that will produce equivalent behavior
    """
    agg_type = fl_aggregator_config._target_
    # if FL global optimizer was Adam, then FL local lr should be 1.0,
    # otherwise we cannot produce same results between FL and non-FL
    assert ("FedAdam" not in agg_type) or (
        fl_local_lr == 1.0
    ), f"When using FedAdam, fl_local lr should be 1. Instead, its {fl_local_lr}"
    # non_fl lr = fl_global_optimizer_lr*fl_local_optimizer_lr
    # Example: fl_local_optimizer lr=1.0, then non_fl LR = fl_global_optimizer_lr
    # Example: fl_local_optimizer lr=0.1, then non_fl LR = 0.1*fl_global_optimizer_lr
    # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `lr`.
    nonfl_lr = fl_aggregator_config.lr * fl_local_lr

    if "FedAvgWithLR" in agg_type:
        optimizer = torch.optim.SGD(
            nonfl_model.fl_get_module().parameters(),
            lr=nonfl_lr,
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `momentum`.
            momentum=fl_aggregator_config.momentum,
        )
    elif "FedAdam" in agg_type:
        optimizer = torch.optim.Adam(
            nonfl_model.fl_get_module().parameters(),
            lr=nonfl_lr,
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `weight_decay`.
            weight_decay=fl_aggregator_config.weight_decay,
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `beta1`.
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `beta2`.
            betas=(fl_aggregator_config.beta1, fl_aggregator_config.beta2),
            # pyre-fixme[16]: `AsyncAggregatorConfig` has no attribute `eps`.
            eps=fl_aggregator_config.eps,
        )
    else:
        raise AssertionError(f"Unknown aggregator {agg_type}")
    return optimizer


def assert_fl_nonfl_same(
    global_model: IFLModel,
    fl_data_provider: IFLDataProvider,
    nonfl_data_loader: torch.utils.data.DataLoader,
    epochs: int,
    local_lr: float,
    aggregator_config: AsyncAggregatorConfig,
    training_rate=1,
    training_duration_mean=0,
    training_duration_sd=0,
) -> str:
    """
    Given:
        data_for_fl={user1:batch1, user2:batch2}
        data_for_non_fl={batch1, batch2}
    Check that the following produce the same trained model:
    1. FL training, 1 user per round. global_opt=SGD, local_lr=x, global_lr=x
    2. Non-FL training, opt=SGD, lr=x
    Return value: model parameters that don't match between fl and non-fl training
    """
    # will be used later to verify training indeed took place
    reference_untrained_model = copy.deepcopy(global_model)
    nonfl_model = copy.deepcopy(reference_untrained_model)
    nonfl_optimizer = get_nonfl_optimizer(
        nonfl_model=nonfl_model,
        fl_local_lr=local_lr,
        fl_aggregator_config=aggregator_config,
    )
    trained_fl_model, trained_nonfl_model = run_fl_nonfl_training(
        fl_model=global_model,
        nonfl_model=nonfl_model,
        nonfl_optimizer=nonfl_optimizer,
        fl_data_provider=fl_data_provider,
        nonfl_data_loader=nonfl_data_loader,
        epochs=epochs,
        fl_local_lr=local_lr,
        fl_aggregator_config=aggregator_config,
        training_rate=training_rate,
        training_duration_mean=training_duration_mean,
        training_duration_sd=training_duration_sd,
    )
    return verify_models_equivalent_after_training(
        trained_fl_model,
        trained_nonfl_model,
        reference_untrained_model,
        rel_epsilon=1e-4,
        abs_epsilon=1e-4,
    )


def run_fl_nonfl_training(
    fl_model: IFLModel,
    nonfl_model: IFLModel,
    nonfl_optimizer: torch.optim.Optimizer,
    fl_data_provider: IFLDataProvider,
    nonfl_data_loader: torch.utils.data.DataLoader,
    epochs: int,
    fl_local_lr: float,
    fl_aggregator_config: AsyncAggregatorConfig,
    training_rate=1,
    training_duration_mean=0,
    training_duration_sd=0,
    example_weight_config: AsyncExampleWeightConfig = EQUAL_EXAMPLE_WEIGHT_CONFIG,
    staleness_weight_config: AsyncStalenessWeightConfig = CONSTANT_STALENESS_WEIGHT_CONFIG,
) -> Tuple[IFLModel, IFLModel]:
    """Run the following training
    1. FL training: train fl_model with fl_data_provider, fl_aggregator_config and fl_local_lr,
    2. Non-FL training: train nonfl_model with nonfl_data_loader and nonfl_optmizer
    """
    fl_trained_model, _ = run_fl_training(
        fl_model=fl_model,
        fl_data_provider=fl_data_provider,
        epochs=epochs,
        local_lr=fl_local_lr,
        aggregator_config=fl_aggregator_config,
        training_rate=training_rate,
        training_duration_mean=training_duration_mean,
        training_duration_sd=training_duration_sd,
        example_weight_config=example_weight_config,
        staleness_weight_config=staleness_weight_config,
        always_keep_trained_model=True,
    )
    nonfl_trained_model = FLTestUtils.run_nonfl_training(
        model=nonfl_model,
        optimizer=nonfl_optimizer,
        data_loader=nonfl_data_loader,
        epochs=epochs,
    )
    return fl_trained_model, nonfl_trained_model


def run_fl_training(
    fl_model: IFLModel,
    fl_data_provider: IFLDataProvider,
    epochs: int,
    local_lr: float,
    aggregator_config: AsyncAggregatorConfig,
    training_rate=1,
    training_duration_mean=0,
    training_duration_sd=0,
    example_weight_config: AsyncExampleWeightConfig = EQUAL_EXAMPLE_WEIGHT_CONFIG,
    staleness_weight_config: AsyncStalenessWeightConfig = CONSTANT_STALENESS_WEIGHT_CONFIG,
    metrics_reporter: Optional[IFLMetricsReporter] = None,
    do_eval: bool = False,
    report_train_metrics_after_aggregation: bool = False,
    eval_epoch_frequency: float = 1.0,
    always_keep_trained_model: float = False,
) -> Tuple[IFLModel, Any]:
    torch.manual_seed(1)
    async_trainer = create_async_trainer(
        model=fl_model,
        local_lr=local_lr,
        epochs=epochs,
        aggregator_config=aggregator_config,
        event_generator_config=create_event_generator_config(
            training_rate=training_rate,
            training_duration_mean=training_duration_mean,
            training_duration_sd=training_duration_sd,
        ),
        example_weight_config=example_weight_config,
        staleness_weight_config=staleness_weight_config,
        do_eval=do_eval,
        report_train_metrics_after_aggregation=report_train_metrics_after_aggregation,
        eval_epoch_frequency=eval_epoch_frequency,
    )
    if metrics_reporter is None:
        metrics_reporter = FakeMetricReporter()
    fl_trained_model, test_metrics = async_trainer.train(
        data_provider=fl_data_provider,
        metric_reporter=metrics_reporter,
        num_total_users=fl_data_provider.num_users(),
        distributed_world_size=1,
    )
    return fl_trained_model, test_metrics


def get_data(
    num_examples: int,
    num_fl_users: int,
    examples_per_user: int,
    fl_batch_size: int,
    nonfl_batch_size: int,
    model: IFLModel,
) -> Tuple[IFLDataProvider, torch.utils.data.DataLoader]:
    fl_data_provider = get_fl_data_provider(
        num_examples=num_examples,
        num_fl_users=num_fl_users,
        examples_per_user=examples_per_user,
        batch_size=fl_batch_size,
        model=model,
    )
    # non-FL data
    dummy_dataset = DummyAlphabetDataset(num_examples)
    nonfl_data_loader = torch.utils.data.DataLoader(
        dummy_dataset, batch_size=nonfl_batch_size, shuffle=False
    )
    return fl_data_provider, nonfl_data_loader


def get_fl_data_provider(
    num_examples: int,
    num_fl_users: int,
    examples_per_user: int,
    batch_size: int,
    model: IFLModel,
) -> IFLDataProvider:
    dummy_dataset = DummyAlphabetDataset(num_examples)
    data_provider, data_loader = DummyAlphabetDataset.create_data_provider_and_loader(
        dummy_dataset, examples_per_user, batch_size, model
    )
    return data_provider
    assert data_loader.num_total_users == num_fl_users, "Error in data sharding"


def get_nonfl_batch_size(fl_batch_size: int, min_examples_per_user: int):
    # how to chose batch_size for non-fl training?
    # if fl_batch_size is bigger than min_examples_per_user, use fl_batch_size
    # however, if fl_batch_size is *smaller* than min_examples_per_user, must
    # use min_examples_per_user as the batch size.
    # example: fl_batch_size=128. 2 users, each with 4 data points
    # non-fl training must run 2 batches of training to match fl-training.
    # so batch size must be 4
    # example: fl_batch_size=8. min_examples_per_user = 1
    # nonfl_batch_size = 8
    return min(fl_batch_size, min_examples_per_user)


def _equal_data_split_params(
    num_examples: int,
    num_fl_users: int,
    fl_batch_size: int,
    one_batch_per_user_only: bool,
) -> Tuple[int, int]:
    """Assume FL data is equally split among users
    Find examples_per_user and nonfl_batch_size
    """
    assert (
        num_examples % num_fl_users == 0
    ), f"Expect num_examples({num_examples}) to be multiple of num_fl_users({num_fl_users})"
    examples_per_user = num_examples // num_fl_users
    if one_batch_per_user_only:
        assert (
            fl_batch_size >= examples_per_user
        ), f"Expected each user to have max 1 batch. Batch_size:{fl_batch_size}, num_examples_per_user:{examples_per_user}"
    nonfl_batch_size = get_nonfl_batch_size(
        fl_batch_size=fl_batch_size, min_examples_per_user=examples_per_user
    )
    assert nonfl_batch_size == min(
        fl_batch_size, examples_per_user
    ), f"fl_batch_size:{fl_batch_size}, nonfl_batch_size:{nonfl_batch_size} "
    f"examples_per_user:{examples_per_user}. To ensure FL and non-FL take the same number of local SGD steps,"
    "nonfl_batch_size should be the same as the lower of (examples_per_user, fl_batch_size)"

    return examples_per_user, nonfl_batch_size


def _unequal_data_split_params(
    num_examples: int,
    num_fl_users: int,
    max_examples_per_user: int,
    fl_batch_size: int,
    one_batch_per_user_only: bool,
) -> Tuple[int, int]:
    r"""
    FL data may be unequally split among users with sequential sharding
    E.g: max_num_examples_per_user = 8. total examples = 12. num_users=2
    user1: 8 examples. user2: 4 examples
    """
    # if equal split. i.e, if every user can get max_examples_per_user
    if num_examples / num_fl_users == max_examples_per_user:
        return _equal_data_split_params(
            num_examples=num_examples,
            num_fl_users=num_fl_users,
            fl_batch_size=fl_batch_size,
            one_batch_per_user_only=one_batch_per_user_only,
        )
    # we must have enough examples such that at least one user gets max_examples_per_user
    assert max_examples_per_user > num_fl_users
    # last user gets leftover examples
    examples_with_last_user = num_examples % max_examples_per_user
    if one_batch_per_user_only:
        assert fl_batch_size >= max_examples_per_user, (
            f"Expected each user to have max 1 batch. Batch_size:{fl_batch_size},"
            f" num_examples_per_user:{max_examples_per_user}"
        )
    nonfl_batch_size = get_nonfl_batch_size(
        fl_batch_size=fl_batch_size, min_examples_per_user=examples_with_last_user
    )
    assert nonfl_batch_size == min(fl_batch_size, max_examples_per_user), (
        f"fl_batch_size:{fl_batch_size}, nonfl_batch_size:{nonfl_batch_size} "
        f"max_examples_per_user:{max_examples_per_user}. "
        f"To ensure FL and non-FL take the same number of local SGD steps,"
        f"nonfl_batch_size should be the same as the lower of (max_examples_per_user, fl_batch_size)",
    )
    return max_examples_per_user, nonfl_batch_size


def get_unequal_split_data(
    num_examples: int,
    num_fl_users: int,
    max_examples_per_user: int,
    fl_batch_size: int,
    model: IFLModel,
    one_batch_per_user_only: bool = False,
):
    r"""
    If FL data is unequally split among users with sequential sharding
    E.g: max_num_examples_per_user = 8. total examples = 12. num_users=2
    user1: 8 examples. user2: 4 examples
    Return FLDataProvider and non-FL DataLoader
    """
    examples_per_user, nonfl_batch_size = _unequal_data_split_params(
        num_examples=num_examples,
        num_fl_users=num_fl_users,
        max_examples_per_user=max_examples_per_user,
        fl_batch_size=fl_batch_size,
        one_batch_per_user_only=one_batch_per_user_only,
    )
    return get_data(
        num_examples=num_examples,
        num_fl_users=num_fl_users,
        examples_per_user=examples_per_user,
        fl_batch_size=fl_batch_size,
        nonfl_batch_size=nonfl_batch_size,
        model=model,
    )


def get_equal_split_data(
    num_examples: int,
    num_fl_users: int,
    fl_batch_size: int,
    model: IFLModel,
    one_batch_per_user_only: bool = False,
) -> Tuple[IFLDataProvider, torch.utils.data.DataLoader]:
    """Assume FL data is equally split among users
    Return FLDataProvider and non-fl DataLoader
    """
    examples_per_user, nonfl_batch_size = _equal_data_split_params(
        num_examples=num_examples,
        num_fl_users=num_fl_users,
        fl_batch_size=fl_batch_size,
        one_batch_per_user_only=one_batch_per_user_only,
    )
    return get_data(
        num_examples=num_examples,
        num_fl_users=num_fl_users,
        examples_per_user=examples_per_user,
        fl_batch_size=fl_batch_size,
        nonfl_batch_size=nonfl_batch_size,
        model=model,
    )


def assert_fl_nonfl_same_equal_data_split(
    fl_batch_size: int,
    num_examples: int,
    num_fl_users: int,
    epochs: int,
    local_lr: float,
    aggregator_config: AsyncAggregatorConfig,
    training_rate=1,
    training_duration_mean=0,
    training_duration_sd=0,
) -> str:
    # TODO: can this test share common code (eg, nonFL training) with
    # test_trainer._test_fl_nonfl_same_equal_data_split?
    torch.manual_seed(1)
    # create dummy FL model on alphabet
    global_model = DummyAlphabetFLModel()
    # if aggregator is FedAdam, each user should have only one batch
    # otherwise non-FL and FL training won't be the same
    one_batch_per_user_only = "FedAdam" in aggregator_config._target_
    fl_data_provider, nonfl_data_loader = get_equal_split_data(
        num_examples=num_examples,
        num_fl_users=num_fl_users,
        fl_batch_size=fl_batch_size,
        model=global_model,
        one_batch_per_user_only=one_batch_per_user_only,
    )
    return assert_fl_nonfl_same(
        global_model=global_model,
        fl_data_provider=fl_data_provider,
        nonfl_data_loader=nonfl_data_loader,
        epochs=epochs,
        local_lr=local_lr,
        aggregator_config=aggregator_config,
        training_rate=training_rate,
        training_duration_mean=training_duration_mean,
        training_duration_sd=training_duration_sd,
    )


def async_train_one_user(
    global_model_at_training_start: IFLModel,
    global_model_at_training_end: IFLModel,
    batches,
    local_lr: float,
) -> IFLModel:
    local_model = copy.deepcopy(global_model_at_training_start)
    local_optimizer = torch.optim.SGD(
        local_model.fl_get_module().parameters(), lr=local_lr
    )
    for batch in batches:
        FLTestUtils.run_nonfl_training_one_batch(
            model=local_model, optimizer=local_optimizer, training_batch=batch
        )
    simulate_async_global_model_update(
        global_model=global_model_at_training_end,
        local_model_before_training=global_model_at_training_start,
        local_model_after_training=local_model,
    )
    return global_model_at_training_end


def simulate_async_global_model_update(
    global_model: IFLModel,
    local_model_before_training: IFLModel,
    local_model_after_training: IFLModel,
):
    # TODO: use AsyncAggregator._update_global_model, after John's refactoring
    reconstructed_grad = copy.deepcopy(global_model)
    FLModelParamUtils.reconstruct_gradient(
        old_model=local_model_before_training.fl_get_module(),
        new_model=local_model_after_training.fl_get_module(),
        grads=reconstructed_grad.fl_get_module(),
    )
    FLModelParamUtils.set_gradient(
        model=global_model.fl_get_module(),
        reference_gradient=reconstructed_grad.fl_get_module(),
    )
    optimizer = torch.optim.SGD(global_model.fl_get_module().parameters(), lr=1.0)
    optimizer.step()


def run_fl_training_with_event_generator(
    fl_model: IFLModel,
    fl_data_provider: IFLDataProvider,
    epochs: int,
    local_lr: float,
    aggregator_config: AsyncAggregatorConfig,
    training_event_generator_config: EventGeneratorConfig,
    example_weight_config: AsyncExampleWeightConfig = EQUAL_EXAMPLE_WEIGHT_CONFIG,
    staleness_weight_config: AsyncStalenessWeightConfig = CONSTANT_STALENESS_WEIGHT_CONFIG,
) -> IFLModel:
    torch.manual_seed(1)
    async_trainer = create_async_trainer(
        model=fl_model,
        local_lr=local_lr,
        epochs=epochs,
        aggregator_config=aggregator_config,
        event_generator_config=training_event_generator_config,
        example_weight_config=example_weight_config,
        staleness_weight_config=staleness_weight_config,
    )
    fl_trained_model, _ = async_trainer.train(
        data_provider=fl_data_provider,
        metric_reporter=FakeMetricReporter(),
        num_total_users=fl_data_provider.num_users(),
        distributed_world_size=1,
    )
    return fl_trained_model


def get_safe_global_lr(fl_batch_size: int, max_examples_per_user: int) -> float:
    """Return a global_lr to use in FL, that can produce equivalent training
    results as non-fl training.
    Return value: either 1.0, or a random float between 0 and 10
    global_lr can be anything if each user has exactly 1 batch.
    otherwise, it must be 1.0
    why? because FL will take exactly one global step for each user
    while non-FL will take num_examples/batch_size global steps
    """
    if fl_batch_size >= max_examples_per_user:
        return np.random.random_sample() * 10
    else:
        return 1.0



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

"""Contains helper for Python PyTest tests.

There are a list of helper utilities missing in PyTest.
We define some of the helpful utilities in this one file
so that all FL-sim PyTests can reuse them easily.
"""
from typing import Any, Sized, List


def assertIsInstance(o: object, t: type) -> None:
    assert isinstance(o, t)


def assertAlmostEqual(first, second, places=None, msg=None, delta=None):
    """Assert that ``first`` and ``second`` is almost equal to each other.

    The equality of ``first`` and ``second`` is determined in a similar way to
    the ``assertAlmostEqual`` function of the standard library.
    """
    if delta is not None and places is not None:
        raise TypeError("specify delta or places - not both")

    msg: str = ""
    diff = abs(second - first)
    if delta is not None:
        if diff <= delta:
            return

        msg = (
            f"The difference between {first} and {second} is not within {delta} delta."
        )
    else:
        if places is None:
            places = 7

        if round(diff, places) == 0:
            return

        msg = f"The difference between {first} and {second} is not within {places} decimal places."

    raise AssertionError(msg)


def assertEqual(object1: object, object2: object, e: object = None) -> None:
    assert object1 == object2, e


def assertNotEqual(object1: object, object2: object, e: object = None) -> None:
    assert object1 != object2, e


def assertLess(object1, object2) -> None:
    assert object1 < object2


def assertLessEqual(object1, object2) -> None:
    assert object1 <= object2


def assertGreaterEqual(object1, object2) -> None:
    assert object1 >= object2


def assertIsNotNone(o: object) -> None:
    assert o is not None


def assertTrue(o: object, e: object = None) -> None:
    assert o, e


def assertFalse(o: object, e: object = None) -> None:
    assert not o, e


def assertEmpty(o: Sized, msg: object = None) -> None:
    assert len(o) == 0, msg


def assertNotEmpty(o: Sized, msg: object = None) -> None:
    assert len(o) > 0, msg


def assertListEqual(l1: List, l2: List) -> None:
    assert l1 == l2


class assertRaises(object):
    def __init__(self, expected_exc: type) -> None:
        self.expected_exc = expected_exc

    def __enter__(self) -> "assertRaises":
        return self

    def __exit__(self, exc_type: type, exc_value: Exception, traceback: Any) -> bool:
        if exc_type is None:
            raise AssertionError(f"{self.expected_exc} was not raised")
        return isinstance(exc_value, self.expected_exc)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from flsim.utils.config_utils import _flatten_dict, fl_json_to_dotlist
from libfb.py import testutil


class ConfigUtilsTest(testutil.BaseFacebookTestCase):
    def setUp(self) -> None:
        super().setUp()

    def test_flatten_dict(self):
        self.assertEqual(_flatten_dict({}), {})
        self.assertEqual(_flatten_dict({"a": 1}), {"a": 1})
        self.assertEqual(_flatten_dict({"a": None}), {"a": None})

        # checks neesting, exp notation
        self.assertEqual(
            _flatten_dict(
                {
                    "a": {
                        "b": {"c": {"val": 3}, "val": 2, "_base_": "b"},
                        "_base_": "a",
                        "val": 1,
                    },
                    "d": "1e-4",
                }
            ),
            {
                "a.b.c.val": 3,
                "a.b.val": 2,
                "a.b._base_": "b",
                "a._base_": "a",
                "a.val": 1,
                "d": "1e-4",
            },
        )

        # checks string floats
        self.assertEqual(_flatten_dict({"e": "5.5"}), {"e": '"5.5"'})

        # make sure json in list remains untouched
        self.assertEqual(
            _flatten_dict(
                {
                    "a": {"b": 1},
                    "l": [1, 2, 3],
                    "ld": [{"a": 1, "b": {"bb": 2}, "c": [11, 22]}, {"z": "xyz"}],
                }
            ),
            {
                "a.b": 1,
                "l": [1, 2, 3],
                "ld": [{"a": 1, "b": {"bb": 2}, "c": [11, 22]}, {"z": "xyz"}],
            },
        )

        # make sure json in key with suffix _dict remains untouched
        self.assertEqual(
            _flatten_dict(
                {
                    "a": {"b": 1},
                    "c": {"d_dict": {"A": 1, "B": "2.2", "C": {"key": "three"}}},
                }
            ),
            {
                "a.b": 1,
                "c.d_dict": {"A": 1, "B": "2.2", "C": {"key": "three"}},
            },
        )

        # check with _base_
        self.assertEqual(
            _flatten_dict(
                {
                    "_base_": {"_base_": "base1", "_base": "base2", "base_": "base3"},
                    "_base": {"_base_": "base1", "_base": "base2", "base_": "base3"},
                    "base_": {"_base_": "base1", "_base": "base2", "base_": "base3"},
                }
            ),
            {
                "_base_._base_": "base1",
                "_base_._base": "base2",
                "_base_.base_": "base3",
                "_base._base_": "base1",
                "_base._base": "base2",
                "_base.base_": "base3",
                "base_._base_": "base1",
                "base_._base": "base2",
                "base_.base_": "base3",
            },
        )

    def test_json_to_dotlist(self):
        self.assertEqual(fl_json_to_dotlist({}, append_or_override=False), [])
        self.assertEqual(
            fl_json_to_dotlist({"a": 1}, append_or_override=False), ["a=1"]
        )
        self.assertEqual(
            fl_json_to_dotlist({"a": None}, append_or_override=False), ["a=null"]
        )

        # checks neesting, exp notation
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "a": {
                        "b": {"c": {"val": 3}, "val": 2, "_base_": "b"},
                        "_base_": "a",
                        "val": 1,
                    },
                    "d": "1e-4",
                },
                append_or_override=False,
            ),
            ["+a@a=a", "+b@a.b=b", "d=1e-4", "a.val=1", "a.b.val=2", "a.b.c.val=3"],
        )

        # checks string floats
        self.assertEqual(
            fl_json_to_dotlist({"e": "5.5"}, append_or_override=False), ['e="5.5"']
        )

        # make sure json in list remains untouched
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "a": {"b": 1},
                    "l": [1, 2, 3],
                    "ld": [{"a": 1, "b": {"bb": 2}, "c": [11, 22]}, {"z": "xyz"}],
                },
                append_or_override=False,
            ),
            [
                "l=[1, 2, 3]",
                "ld=[{'a': 1, 'b': {'bb': 2}, 'c': [11, 22]}, {'z': 'xyz'}]",
                "a.b=1",
            ],
        )

        # make sure json in key with suffix _dict is handled correctly
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "a": {"b": 1},
                    "c": {"d_dict": {"A": 1, "B": "2.2", "C": {"3": "three"}}},
                },
                append_or_override=False,
            ),
            [
                "a.b=1",
                'c.d_dict="{\\"A\\": 1, \\"B\\": \\"2.2\\", \\"C\\": {\\"3\\": \\"three\\"}}"',
            ],
        )

        # check with _base_
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "_base_": {
                        "_base_": "base1",
                        "_base": "base2",
                        "base_": "base3",
                        "base": "just_base",
                    },
                    "_base": {
                        "_base_": "base1",
                        "_base": "base2",
                        "base_": "base3",
                        "base": "just_base",
                    },
                    "base_": {
                        "_base_": "base1",
                        "_base": "base2",
                        "base_": "base3",
                        "base": "just_base",
                    },
                    "base": "just_base",
                },
                append_or_override=False,
            ),
            [
                "+_base@_base=base1",
                "+_base_@_base_=base1",
                "+base_@base_=base1",
                "base=just_base",
                "_base._base=base2",
                "_base.base=just_base",
                "_base.base_=base3",
                "_base_._base=base2",
                "_base_.base=just_base",
                "_base_.base_=base3",
                "base_._base=base2",
                "base_.base=just_base",
                "base_.base_=base3",
            ],
        )

    def test_json_to_dotlist_append_or_override(self):

        self.assertEqual(fl_json_to_dotlist({}), [])
        self.assertEqual(fl_json_to_dotlist({"a": 1}), ["++a=1"])
        self.assertEqual(fl_json_to_dotlist({"a": None}), ["++a=null"])

        # checks neesting, exp notation
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "a": {
                        "b": {"c": {"val": 3}, "val": 2, "_base_": "b"},
                        "_base_": "a",
                        "val": 1,
                    },
                    "d": "1e-4",
                }
            ),
            [
                "+a@a=a",
                "+b@a.b=b",
                "++d=1e-4",
                "++a.val=1",
                "++a.b.val=2",
                "++a.b.c.val=3",
            ],
        )

        # checks string floats
        self.assertEqual(fl_json_to_dotlist({"e": "5.5"}), ['++e="5.5"'])

        # make sure json in list remains untouched
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "a": {"b": 1},
                    "l": [1, 2, 3],
                    "ld": [{"a": 1, "b": {"bb": 2}, "c": [11, 22]}, {"z": "xyz"}],
                }
            ),
            [
                "++l=[1, 2, 3]",
                "++ld=[{'a': 1, 'b': {'bb': 2}, 'c': [11, 22]}, {'z': 'xyz'}]",
                "++a.b=1",
            ],
        )

        # make sure json in key with suffix _dict is handled correctly
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "a": {"b": 1},
                    "c": {"d_dict": {"A": 1, "B": "2.2", "C": {"3": "three"}}},
                }
            ),
            [
                "++a.b=1",
                '++c.d_dict="{\\"A\\": 1, \\"B\\": \\"2.2\\", \\"C\\": {\\"3\\": \\"three\\"}}"',
            ],
        )

        # check with _base_
        self.assertEqual(
            fl_json_to_dotlist(
                {
                    "_base_": {
                        "_base_": "base1",
                        "_base": "base2",
                        "base_": "base3",
                        "base": "just_base",
                    },
                    "_base": {
                        "_base_": "base1",
                        "_base": "base2",
                        "base_": "base3",
                        "base": "just_base",
                    },
                    "base_": {
                        "_base_": "base1",
                        "_base": "base2",
                        "base_": "base3",
                        "base": "just_base",
                    },
                    "base": "just_base",
                }
            ),
            [
                "+_base@_base=base1",
                "+_base_@_base_=base1",
                "+base_@base_=base1",
                "++base=just_base",
                "++_base._base=base2",
                "++_base.base=just_base",
                "++_base.base_=base3",
                "++_base_._base=base2",
                "++_base_.base=just_base",
                "++_base_.base_=base3",
                "++base_._base=base2",
                "++base_.base=just_base",
                "++base_.base_=base3",
            ],
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import List

import pytest
from flsim.clients.async_client import AsyncClientDevice
from flsim.clients.base_client import Client, ClientConfig
from flsim.common.pytest_helper import assertEqual
from flsim.common.training_event_handler import TestAsyncTrainingEventHandler
from flsim.data.data_provider import FLDataProviderFromList, FLUserDataFromList
from flsim.utils.async_trainer.async_user_selector import AsyncUserSelectorInfo
from flsim.utils.async_trainer.device_state import TrainingSchedule
from flsim.utils.sample_model import MockFLModel
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from omegaconf import OmegaConf


@pytest.fixture(scope="class")
def prepare_job_scheduler_util_test(request):
    request.cls.shared_client_config = ClientConfig(
        epochs=1,
        max_clip_norm_normalized=0,
        only_federated_params=True,
        random_seed=1,
        store_models_and_optimizers=False,
    )


@pytest.mark.usefixtures("prepare_job_scheduler_util_test")
class TestJobSchedulerUtil:
    def _build_data_provider(
        self, num_examples, examples_per_user, user_batch_size, global_model
    ) -> FLDataProviderFromList:
        dummy_dataset = DummyAlphabetDataset(num_examples)
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, examples_per_user, user_batch_size, global_model
        )
        return data_provider

    def _create_mock_base_client_info(self, examples_per_user, user_index):
        user_data = FLUserDataFromList(
            [1] * examples_per_user,
            MockFLModel(num_examples_per_user=examples_per_user),
        )
        user_info = AsyncUserSelectorInfo(user_data=user_data, user_index=user_index)
        base_client = Client(
            **OmegaConf.structured(self.shared_client_config),
            dataset=user_info.user_data,
            name=f"client_{user_info.user_index}",
        )
        return user_info, base_client

    def test_test_async_job_scheduler(self) -> None:
        num_users = 100
        model_trainer = TestAsyncTrainingEventHandler()
        assertEqual(len(model_trainer.num_unseen_global_model_updates), 0)
        assertEqual(len(model_trainer.training_events), 0)

        for user_index in range(num_users):
            training_schedule = TrainingSchedule(
                creation_time=0,
                start_time=user_index,
                end_time=num_users + (num_users - user_index - 1),
            )
            user_info, base_client = self._create_mock_base_client_info(
                examples_per_user=1, user_index=user_index
            )

            client = AsyncClientDevice(
                training_schedule=training_schedule,
                client=base_client,
                user_info=user_info,
            )
            model_trainer.on_training_start(client)
            assertEqual(client.model_seqnum, model_trainer.current_seqnum)
            assertEqual(client.model_seqnum, user_index)
            model_trainer.on_training_end(client)
        # each user generates two events: one TrainingFinished, and one TrainingStart
        assertEqual(len(model_trainer.training_events), 2 * num_users)
        # training proceeds sequentially, so num_unseen_global_model_updates should
        # be a list of zeroes
        assertEqual(model_trainer.num_unseen_global_model_updates, [0] * num_users)

    def test_staleness_compute_in_job_scheduler(self) -> None:
        # train in LIFO order: the first user to start training is the last
        # one to finish training
        # in this case, assume n users, observed staleness should be:
        # 0 (user n), 1 (user n-1), 2 (user n-2)... n (user 1)
        # timeline:
        # USERS START TRAINING
        # t=0, user 0 starts training
        # t=1, user 1 starts training
        # ...
        # t=n-1, user n-1 starts training
        # USERS END TRAINING
        # t=n, user n-1 ends training
        # t=n+1, user n-2 ends training
        # ...
        # t=n+(n-1), user 0 ends training

        num_users = 100
        model_trainer = TestAsyncTrainingEventHandler()
        clients: List[AsyncClientDevice] = []
        for user_index in range(num_users):
            # user i starts training at time i, ends training at time n + (n-i-1)
            training_schedule = TrainingSchedule(
                creation_time=0,
                start_time=user_index,
                end_time=num_users + (num_users - user_index - 1),
            )
            user_info, base_client = self._create_mock_base_client_info(
                examples_per_user=1, user_index=user_index
            )

            client = AsyncClientDevice(
                training_schedule=training_schedule,
                client=base_client,
                user_info=user_info,
            )
            model_trainer.on_training_start(client)
            clients.append(client)
        while len(clients):
            client = clients.pop()
            model_trainer.on_training_end(client)
        # observed staleness should be [0, 1, 2, ....num_users-1]
        assertEqual(
            model_trainer.num_unseen_global_model_updates, list(range(num_users))
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from enum import IntEnum

import torch
import torch.nn as nn
from flsim.utils.distributed.fl_distributed import FLDistributedUtils, OperationType
from flsim.utils.fl.common import FLModelParamUtils


class AggregationType(IntEnum):
    """
    Type of averaging for the aggregator
    """

    AVERAGE = 0
    SUM = 1
    WEIGHTED_AVERAGE = 2
    WEIGHTED_SUM = 3


class Aggregator:
    """
    Util class to handle aggregation logic such as
    {weighted, unweighted}_summation, {weighted, unweighted}_averaging

    Please do not extend this class
    """

    def __init__(
        self,
        module: nn.Module,
        aggregation_type: AggregationType,
        only_federated_params: bool = True,
    ):
        self._buffer_module: nn.Module = FLModelParamUtils.clone(module)
        self.device = next(self._buffer_module.parameters()).device
        self._sum_weights: torch.Tensor = torch.zeros(1, device=self.device)
        self.only_federated_params = only_federated_params
        FLModelParamUtils.zero_weights(
            self._buffer_module, only_federated_params=self.only_federated_params
        )
        self.aggregation_type = aggregation_type

    def zero_weights(self):
        FLModelParamUtils.zero_weights(
            self._buffer_module, only_federated_params=self.only_federated_params
        )
        self._sum_weights = torch.zeros(1, device=self.device)

    def add_update(self, delta: nn.Module, weight: float):
        weight = weight if self._is_weighted else 1.0
        FLModelParamUtils.add_model(delta, self._buffer_module, self._buffer_module)
        self._sum_weights += weight

    def apply_weight_to_update(self, delta: nn.Module, weight: float):
        weight = weight if self._is_weighted else 1.0
        FLModelParamUtils.multiply_model_by_weight(
            model=delta,
            weight=weight,
            model_to_save=delta,
        )

    def aggregate(
        self, distributed_op: OperationType = OperationType.SUM_AND_BROADCAST
    ) -> nn.Module:
        FLDistributedUtils.synchronize_across_ranks(
            model=self._buffer_module,
            weights=self._sum_weights,
            operation=distributed_op,
        )

        if self._is_averaged:
            FLModelParamUtils.multiply_model_by_weight(
                model=self._buffer_module,
                weight=1.0 / self.sum_weights.item(),
                model_to_save=self._buffer_module,
            )
        return self._buffer_module

    @property
    def sum_weights(self) -> torch.Tensor:
        return self._sum_weights

    @property
    def _is_weighted(self) -> bool:
        return self.aggregation_type in [
            AggregationType.WEIGHTED_AVERAGE,
            AggregationType.WEIGHTED_SUM,
        ]

    @property
    def _is_averaged(self) -> bool:
        return self.aggregation_type in [
            AggregationType.WEIGHTED_AVERAGE,
            AggregationType.AVERAGE,
        ]



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from flsim.interfaces.model import IFLModel
from flsim.utils.simple_batch_metrics import FLBatchMetrics
from torch import Tensor


class TestDataSetting:
    NUM_SHARDS: int = 10
    SHARDING_COL_INDEX: int = 2
    TEXT_COL_NAME: str = "text"
    LABEL_COL_NAME: str = "label"
    USER_ID_COL_NAME: str = "user_id"


class SimpleLinearNet(nn.Module):
    def __init__(self, D_in: int, D_out: int) -> None:
        """
        We create a simple linear model, with input dimension D_in and output dimension D_out
        """
        super(SimpleLinearNet, self).__init__()
        self.linear = nn.Linear(in_features=D_in, out_features=D_out, bias=False)

    def forward(self, x) -> Tensor:
        return self.linear(x)


class LinearFLModel(IFLModel):
    def __init__(
        self, D_in: int = 40, D_out: int = 1, use_cuda_if_available: bool = False
    ) -> None:
        """
        create a sample dummy FL model for alphabet dataset
        """
        self.model = SimpleLinearNet(D_in, D_out)
        self.use_cuda_if_available = use_cuda_if_available

    def fl_forward(self, batch) -> FLBatchMetrics:
        text = batch[TestDataSetting.TEXT_COL_NAME]
        batch_label = batch[TestDataSetting.LABEL_COL_NAME]
        # stacked_label = torch.tensor(batch_label.view(-1), dtype=torch.long)
        stacked_label = batch_label

        if self.use_cuda_if_available:
            text = text.cuda()

        out = self.model(text)
        if self.use_cuda_if_available:
            out, batch_label, stacked_label = (
                out.cuda(),
                batch[TestDataSetting.LABEL_COL_NAME].cuda(),
                stacked_label.cuda(),
            )

        loss = F.mse_loss(out, stacked_label)
        # loss = F.mse_loss(out, batch_label)
        num_examples = self.get_num_examples(batch)
        return FLBatchMetrics(
            loss=loss,
            num_examples=num_examples,
            predictions=out,
            targets=batch_label,
            model_inputs=batch,
        )

    def fl_create_training_batch(self, **kwargs) -> None:
        return kwargs.get("batch", None)

    def fl_get_module(self) -> nn.Module:
        return self.model

    def fl_cuda(self) -> None:
        self.model = self.model.cuda()

    def get_eval_metrics(self, batch) -> FLBatchMetrics:
        with torch.no_grad():
            return self.fl_forward(batch)

    def get_num_examples(self, batch) -> int:
        return len(batch[TestDataSetting.LABEL_COL_NAME])


class TwoLayerNet(nn.Module):
    def __init__(self, D_in: int, H: int, D_out: int) -> None:
        """
        In the constructor we instantiate two nn.Linear modules and assign them as
        member variables.

        D_in: input dimension
        H: dimension of hidden layer
        D_out: output dimension
        """
        super(TwoLayerNet, self).__init__()
        self.linear1 = nn.Linear(D_in, H)
        self.linear2 = nn.Linear(H, D_out)

    def forward(self, x) -> Tensor:
        """
        In the forward function we accept a Variable of input data and we must
        return a Variable of output data. We can use Modules defined in the
        constructor as well as arbitrary operators on Variables.
        """
        h_relu = F.relu(self.linear1(x))
        y_pred = self.linear2(h_relu)
        return F.log_softmax(y_pred, 1)


class DummyAlphabetFLModel(IFLModel):
    def __init__(
        self,
        embedding_size: int = 10,
        hidden_dim: int = 8,
        use_cuda_if_available: bool = False,
    ) -> None:
        """
        create a sample dummy FL model for alphabet dataset
        """
        self.model = TwoLayerNet(embedding_size, hidden_dim, 2)
        self.dummy_embedding = torch.rand(26, embedding_size)
        self.use_cuda_if_available = use_cuda_if_available

    def fl_forward(self, batch) -> FLBatchMetrics:
        text = batch[TestDataSetting.TEXT_COL_NAME]
        batch_label = batch[TestDataSetting.LABEL_COL_NAME]
        stacked_label = torch.tensor(batch_label.view(-1), dtype=torch.long)
        text_embeddings = self.dummy_embedding[text, :]

        if self.use_cuda_if_available:
            text_embeddings = text_embeddings.cuda()

        out = self.model(text_embeddings)
        if self.use_cuda_if_available:
            out, batch_label, stacked_label = (
                out.cuda(),
                batch[TestDataSetting.LABEL_COL_NAME].cuda(),
                stacked_label.cuda(),
            )

        loss = F.nll_loss(out, stacked_label)
        # produce a large loss, so gradients are large
        # this prevents unit tests from failing because of numerical issues
        loss.mul_(100.0)
        num_examples = self.get_num_examples(batch)
        return FLBatchMetrics(
            loss=loss,
            num_examples=num_examples,
            predictions=out,
            targets=batch_label,
            model_inputs=text_embeddings,
        )

    def fl_create_training_batch(self, **kwargs) -> None:
        return kwargs.get("batch", None)

    def fl_get_module(self) -> nn.Module:
        return self.model

    def fl_cuda(self) -> None:
        self.model = self.model.cuda()

    def get_eval_metrics(self, batch) -> FLBatchMetrics:
        with torch.no_grad():
            return self.fl_forward(batch)

    def get_num_examples(self, batch) -> int:
        return len(batch[TestDataSetting.LABEL_COL_NAME])


class MockFLModel(IFLModel):
    r"""
    Mock IFLModel for testing that will return
    whatever the user pass into the constructor
    """

    def __init__(
        self,
        num_examples_per_user: int = 1,
        batch_labels: Optional[torch.Tensor] = None,
        model_output: Optional[torch.Tensor] = None,
        model_input: Optional[torch.Tensor] = None,
        loss: Optional[torch.Tensor] = None,
    ) -> None:
        self.model = TwoLayerNet(10, 8, 2)
        self.num_examples_per_user = num_examples_per_user

        self.batch_labels = self._get_or_return_dummy_tensor(batch_labels)
        self.model_output = self._get_or_return_dummy_tensor(model_output)
        self.model_input = self._get_or_return_dummy_tensor(model_input)
        self.loss = self._get_or_return_dummy_tensor(loss)

    def fl_forward(self, batch) -> FLBatchMetrics:
        num_examples = self.get_num_examples(batch)
        return FLBatchMetrics(
            loss=self.loss,
            num_examples=num_examples,
            predictions=self.model_output,
            targets=self.batch_labels,
            model_inputs=self.model_input,
        )

    def fl_create_training_batch(self, **kwargs) -> None:
        return kwargs.get("batch", None)

    def fl_get_module(self) -> nn.Module:
        return self.model

    def fl_cuda(self) -> None:
        pass

    def get_eval_metrics(self, batch) -> FLBatchMetrics:
        return FLBatchMetrics(
            loss=self.loss,
            num_examples=self.num_examples_per_user,
            predictions=self.model_output,
            targets=self.batch_labels,
            model_inputs=self.model_input,
        )

    def get_num_examples(self, batch) -> int:
        return self.num_examples_per_user

    def _get_or_return_dummy_tensor(self, data: Optional[torch.Tensor]):
        return data if data is not None else torch.Tensor([1])


class BiasOnly(nn.Module):
    """This module has only a bias term
    It is useful for unit testing because the gradient will be constant.
    """

    def __init__(self):
        super(BiasOnly, self).__init__()
        self.bias = nn.Parameter(torch.zeros(1))

    def forward(self):
        return self.bias


class ConstantGradientFLModel(IFLModel):
    def __init__(
        self,
    ) -> None:
        """A dummy model where the gradient is constant
        This is useful in testing because the parameters of the model move by a
        a constant amount in each optimizer.step()
        Initial value of self.model.bias = 0
        If using SGD, Final value of self.model.bias = lr * num_times_optimizer_step()_is_Called
        """
        self.model = BiasOnly()

    def fl_forward(self, batch) -> FLBatchMetrics:
        # self.model() will return the value of bias
        # we want the gradient to be negative, so self.model()['bias'] increase
        # with each optimizer.step
        loss = -1 * self.model()
        return FLBatchMetrics(
            loss=torch.Tensor(loss),
            num_examples=1,
            predictions=torch.Tensor(0),
            targets=torch.Tensor(0),
            model_inputs=torch.Tensor(0),
        )

    def fl_create_training_batch(self, **kwargs) -> None:
        return kwargs.get("batch", None)

    def fl_get_module(self) -> nn.Module:
        return self.model

    def fl_cuda(self) -> None:
        self.model = self.model.cuda()

    def get_eval_metrics(self, batch) -> FLBatchMetrics:
        with torch.no_grad():
            return self.fl_forward(batch)

    def get_num_examples(self, batch) -> int:
        return 1



# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import math
import random

from flsim.common.diversity_metrics import (
    DiversityMetrics,
    DiversityMetricType,
    DiversityStatistics,
)
from flsim.common.pytest_helper import (
    assertEqual,
    assertTrue,
)


class TestDiversityStatistics:
    def test_initialize_diversity_statistics(self):
        n_test_cases = 1000
        num_cohorts_per_test = 10
        std_dev = 500.0
        num_tol = 1e-06

        for _ in range(n_test_cases):
            cohort_diversity_metrics = []
            cohort_gradient_diversities = []
            cohort_orthogonalities = []

            for _ in range(num_cohorts_per_test):
                norm_of_sum = abs(random.normalvariate(0, std_dev))
                sum_of_norms = abs(random.normalvariate(0, std_dev))
                diversity_metrics = DiversityMetrics(norm_of_sum, sum_of_norms)

                cohort_diversity_metrics.append(diversity_metrics)
                cohort_gradient_diversities.append(diversity_metrics.gradient_diversity)
                cohort_orthogonalities.append(diversity_metrics.orthogonality)

            diversity_statistics = DiversityStatistics(cohort_diversity_metrics)
            diversity_statistics_eq = DiversityStatistics(cohort_diversity_metrics)
            assertEqual(diversity_statistics, diversity_statistics_eq)

            assertTrue(
                math.isclose(
                    diversity_statistics.maximum_metric,
                    max(cohort_gradient_diversities),
                    rel_tol=num_tol,
                )
            )
            assertTrue(
                math.isclose(
                    diversity_statistics.minimum_metric,
                    min(cohort_gradient_diversities),
                    rel_tol=num_tol,
                )
            )
            assertTrue(
                math.isclose(
                    diversity_statistics.average_metric,
                    sum(cohort_gradient_diversities) / len(cohort_gradient_diversities),
                    rel_tol=num_tol,
                )
            )

            for metric in cohort_diversity_metrics:
                metric.diversity_metric_type = DiversityMetricType.orthogonality

            diversity_statistics = DiversityStatistics(cohort_diversity_metrics)
            assertTrue(
                math.isclose(
                    diversity_statistics.maximum_metric,
                    max(cohort_orthogonalities),
                    rel_tol=num_tol,
                )
            )
            assertTrue(
                math.isclose(
                    diversity_statistics.minimum_metric,
                    min(cohort_orthogonalities),
                    rel_tol=num_tol,
                )
            )
            assertTrue(
                math.isclose(
                    diversity_statistics.average_metric,
                    sum(cohort_orthogonalities) / len(cohort_orthogonalities),
                    rel_tol=num_tol,
                )
            )


class TestDiversityMetrics:
    def test_initialize_diversity_metrics(self):
        diversity_metrics = DiversityMetrics(norm_of_sum=1.0, sum_of_norms=2.0)
        assertTrue(isinstance(diversity_metrics, DiversityMetrics))
        assertEqual(
            diversity_metrics.diversity_metric_type,
            DiversityMetricType.gradient_diversity,
        )

        # Test all comparators for both metrics of interest
        diversity_metrics_incr = DiversityMetrics(norm_of_sum=1.0, sum_of_norms=3.0)
        assertTrue(diversity_metrics_incr > diversity_metrics)
        assertTrue(diversity_metrics_incr >= diversity_metrics)
        assertTrue(diversity_metrics < diversity_metrics_incr)
        assertTrue(diversity_metrics <= diversity_metrics_incr)
        assertTrue(diversity_metrics != diversity_metrics_incr)
        assertTrue(not (diversity_metrics == diversity_metrics_incr))
        diversity_metrics.diversity_metric_type = DiversityMetricType.orthogonality
        diversity_metrics_incr.diversity_metric_type = DiversityMetricType.orthogonality
        assertTrue(diversity_metrics_incr > diversity_metrics)
        assertTrue(diversity_metrics_incr >= diversity_metrics)
        assertTrue(diversity_metrics < diversity_metrics_incr)
        assertTrue(diversity_metrics <= diversity_metrics_incr)
        assertTrue(diversity_metrics != diversity_metrics_incr)
        assertTrue(not (diversity_metrics == diversity_metrics_incr))

        n_test_cases = 1000
        std_dev = 500.0
        numerical_tol = 1e-04
        for _ in range(n_test_cases):
            norm_of_sum = abs(random.normalvariate(0, std_dev))
            sum_of_norms = abs(random.normalvariate(0, std_dev))
            diversity_metrics = DiversityMetrics(norm_of_sum, sum_of_norms)

            # Check bounds on and relationship between the two stored quantities
            assertTrue(diversity_metrics._recpr_gradient_diversity >= 0.0)
            assertTrue(diversity_metrics._recpr_orthogonality >= -1.0)
            assertTrue(
                diversity_metrics.orthogonality < 0
                or diversity_metrics.gradient_diversity
                <= diversity_metrics.orthogonality
            )
            assertTrue(
                math.isclose(
                    diversity_metrics._recpr_gradient_diversity
                    - diversity_metrics._recpr_orthogonality,
                    1.0,
                    rel_tol=numerical_tol,
                )
            )
            diversity_metrics_cpy = copy.deepcopy(diversity_metrics)
            diversity_metrics_eq = DiversityMetrics(norm_of_sum, sum_of_norms)
            assertEqual(diversity_metrics_cpy, diversity_metrics)
            assertEqual(diversity_metrics_eq, diversity_metrics)

            # Increasing sum_of_norms will increase both gradient diversity and orthogonality.
            # Increasing norm_of_sum will decrease them both.
            diversity_metrics_incr = DiversityMetrics(norm_of_sum, sum_of_norms * 1.01)
            diversity_metrics_decr = DiversityMetrics(norm_of_sum * 1.01, sum_of_norms)
            assertTrue(diversity_metrics_incr > diversity_metrics)
            assertTrue(diversity_metrics_decr < diversity_metrics)
            diversity_metrics.diversity_metric_type = DiversityMetricType.orthogonality
            diversity_metrics_incr.diversity_metric_type = (
                DiversityMetricType.orthogonality
            )
            diversity_metrics_decr.diversity_metric_type = (
                DiversityMetricType.orthogonality
            )
            assertTrue(diversity_metrics_incr > diversity_metrics)
            assertTrue(diversity_metrics_decr < diversity_metrics)
            diversity_metrics.diversity_metric_type = (
                DiversityMetricType.gradient_diversity
            )

            # Check that metric_value returns the value of the poperty of interest
            diversity_metrics.diversity_metric_type = DiversityMetricType.orthogonality
            assertTrue(
                math.isclose(
                    diversity_metrics.metric_value, diversity_metrics.orthogonality
                )
            )
            diversity_metrics.diversity_metric_type = (
                DiversityMetricType.gradient_diversity
            )
            assertTrue(
                math.isclose(
                    diversity_metrics.metric_value, diversity_metrics.gradient_diversity
                )
            )

            # Check the special case when norm_of_sum = sum_of_norms
            diversity_metrics_eq = DiversityMetrics(sum_of_norms, sum_of_norms)
            assertTrue(
                math.isclose(
                    diversity_metrics_eq._recpr_gradient_diversity,
                    1.0,
                    rel_tol=numerical_tol,
                )
            )
            assertTrue(
                math.isclose(
                    diversity_metrics_eq._recpr_orthogonality,
                    0.0,
                    rel_tol=numerical_tol,
                )
            )
            assertTrue(
                math.isclose(
                    diversity_metrics_eq.gradient_diversity, 1.0, rel_tol=numerical_tol
                )
            )
            assertTrue(math.isnan(diversity_metrics_eq.orthogonality))



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from flsim.common.pytest_helper import assertEqual, assertTrue, assertRaises
from flsim.secure_aggregation.secure_aggregator import (
    FixedPointConverter,
    SecureAggregator,
    FixedPointConfig,
    utility_config_flatter,
)
from flsim.tests import utils
from omegaconf import OmegaConf


class TestSecureAggregator:
    def _create_model(self, model_param_value):
        """
        Creates a two-layer model
        """
        fl_model = utils.SampleNet(utils.TwoFC())
        fl_model.fl_get_module().fill_all(model_param_value)
        return fl_model.fl_get_module()

    def test_fixedpoint_init(self):
        """
        Tests that FixedPointConverter init works correctly
        """

        converter = FixedPointConverter(
            **OmegaConf.structured(FixedPointConfig(num_bytes=2, scaling_factor=1000))
        )
        assertEqual(converter.max_value, 32767)
        assertEqual(converter.min_value, -32768)

        with assertRaises(ValueError):
            converter = FixedPointConverter(
                **OmegaConf.structured(
                    FixedPointConfig(num_bytes=9, scaling_factor=1000)
                )
            )

        with assertRaises(ValueError):
            converter = FixedPointConverter(
                **OmegaConf.structured(
                    FixedPointConfig(num_bytes=3, scaling_factor=-100)
                )
            )

    def test_floating_to_fixedpoint(self):
        """
        Tests whether conversion from floating point to fixed point works
        """

        #  hence minValue = -32768, maxValue = 32767
        converter = FixedPointConverter(
            **OmegaConf.structured(FixedPointConfig(num_bytes=2, scaling_factor=100))
        )

        x = torch.tensor(17.42)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = 1742.0 ==> round to 1742
        assertEqual(y, torch.tensor(1742))

        x = torch.tensor(17.4298)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = 1742.98 ==> round to 1743
        assertEqual(y, torch.tensor(1743))

        x = torch.tensor(-2.34)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = -234.0 ==> round to -234
        assertEqual(y, torch.tensor(-234))

        x = torch.tensor(-2.3456)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = -234.56 ==> round to -235
        assertEqual(y, torch.tensor(-235))

        x = torch.tensor(-2.3416)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = -234.16 ==> round to -234
        assertEqual(y, torch.tensor(-234))

        x = torch.tensor(12345.0167)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = 1234501.67 ==> adjust to maxValue 32767
        assertEqual(y, torch.tensor(32767))

        x = torch.tensor(-327.69)
        y = converter.to_fixedpoint(x)
        # y = x * scaling_factor = -32769 ==> adjust to minValue -32768
        assertEqual(y, torch.tensor(-32768))

    def test_fixed_to_floating_point(self):
        """
        Tests whether conversion from fixed point to floating point works
        """

        converter = FixedPointConverter(
            **OmegaConf.structured(FixedPointConfig(num_bytes=1, scaling_factor=85))
        )

        x = torch.tensor(85)
        y = converter.to_float(x)
        # y = x / scaling_factor = 1.0
        assertTrue(torch.allclose(y, torch.tensor(1.0), rtol=1e-10))

        x = torch.tensor(157)
        y = converter.to_float(x)
        # y = x / scaling_factor = 1.847058823529412
        assertTrue(torch.allclose(y, torch.tensor(1.847058823529412), rtol=1e-10))

    def test_params_floating_to_fixedpoint(self):
        """
        Tests whether the parameters of a model are converted correctly
        from floating point to fixed point
        """

        #  hence minValue = -32768, maxValue = 32767
        config = FixedPointConfig(num_bytes=2, scaling_factor=100)

        model = self._create_model(6.328)
        secure_aggregator = SecureAggregator(utility_config_flatter(model, config))
        secure_aggregator.params_to_fixedpoint(model)
        mismatched = utils.model_parameters_equal_to_value(model, 633.0)
        assertEqual(mismatched, "", mismatched)

        model = self._create_model(-3.8345)
        secure_aggregator = SecureAggregator(utility_config_flatter(model, config))
        secure_aggregator.params_to_fixedpoint(model)
        mismatched = utils.model_parameters_equal_to_value(model, -383.0)
        assertEqual(mismatched, "", mismatched)

    def test_params_floating_to_fixedpoint_different_config_for_layers(self):
        """
        Tests whether the parameters of a model are converted correctly
        from floating point to fixed point, when we have different
        FixedPointConverter configs for different layers

        Notes:
            T97188526
        """

        config_layer1 = FixedPointConfig(num_bytes=2, scaling_factor=100)
        #  hence minValue = -32768, maxValue = 32767
        config_layer2 = FixedPointConfig(num_bytes=1, scaling_factor=10)
        #  hence minValue = -128, maxValue = 127

        config = {}
        config["fc1.weight"] = config_layer1
        config["fc1.bias"] = config_layer1
        config["fc2.weight"] = config_layer2
        config["fc2.bias"] = config_layer2

        model = self._create_model(5.4728)
        secure_aggregator = SecureAggregator(config)
        secure_aggregator.params_to_fixedpoint(model)
        for name, p in model.named_parameters():
            if name == "fc1.weight" or name == "fc1.bias":
                # round 547.28 to 547
                assertTrue(torch.allclose(p, torch.tensor(547.0), rtol=1e-10))
            if name == "fc2.weight" or name == "fc2.bias":
                # round 54.728 to 55
                assertTrue(torch.allclose(p, torch.tensor(55.0), rtol=1e-10))

    def test_error_raised_per_layer_config_not_set(self):
        """
        Tests whether all layers have their corresponding configs, when
        per layer fixed point converter is used.
        """

        config_layer1 = FixedPointConfig(num_bytes=8, scaling_factor=10000)

        config = {}
        config["fc1.weight"] = config_layer1
        config["fc1.bias"] = config_layer1

        model = self._create_model(600)
        secure_aggregator = SecureAggregator(config)

        with assertRaises(ValueError):
            secure_aggregator.params_to_float(model)

        with assertRaises(ValueError):
            secure_aggregator.params_to_fixedpoint(model)

    def test_params_fixed_to_floating_point(self):
        """
        Tests whether the parameters of a model are converted correctly
        from fixed point to floating point
        """
        config = FixedPointConfig(num_bytes=3, scaling_factor=40)
        model = self._create_model(880.0)
        secure_aggregator = SecureAggregator(utility_config_flatter(model, config))
        secure_aggregator.params_to_float(model)
        mismatched = utils.model_parameters_equal_to_value(model, 22.0)
        assertEqual(mismatched, "", mismatched)

    def test_params_fixed_to_floating_point_different_config_for_layers(self):
        """
        Tests whether the parameters of a model are converted correctly
        from fixed point to floating point, when we have different
        FixedPointConverter configs for different layers
        """
        config_layer1 = FixedPointConfig(num_bytes=2, scaling_factor=30)
        config_layer2 = FixedPointConfig(num_bytes=1, scaling_factor=80)

        config = {}
        config["fc1.weight"] = config_layer1
        config["fc1.bias"] = config_layer1
        config["fc2.weight"] = config_layer2
        config["fc2.bias"] = config_layer2

        model = self._create_model(832.8)
        secure_aggregator = SecureAggregator(config)
        secure_aggregator.params_to_float(model)
        for name, p in model.named_parameters():
            if name == "fc1.weight" or name == "fc1.bias":
                # 832.8 / 30 = 27.76
                assertTrue(torch.allclose(p, torch.tensor(27.76), rtol=1e-10))
            if name == "fc2.weight" or name == "fc2.bias":
                # 832.8 / 80 = 10.41
                assertTrue(torch.allclose(p, torch.tensor(10.41), rtol=1e-10))



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from copy import deepcopy
from unittest.mock import MagicMock

import pytest
import torch
import torch.nn as nn
from flsim.clients.base_client import Client, ClientConfig
from flsim.clients.dp_client import DPClient, DPClientConfig
from flsim.common.pytest_helper import (
    assertEqual,
    assertIsInstance,
    assertNotEqual,
    assertTrue,
    assertFalse,
    assertAlmostEqual,
)
from flsim.common.timeout_simulator import (
    GaussianTimeOutSimulator,
    GaussianTimeOutSimulatorConfig,
    NeverTimeOutSimulator,
    NeverTimeOutSimulatorConfig,
)
from flsim.optimizers.local_optimizers import (
    LocalOptimizerSGD,
    LocalOptimizerFedProxConfig,
    LocalOptimizerSGDConfig,
)
from flsim.optimizers.optimizer_scheduler import ConstantLRScheduler
from flsim.optimizers.optimizer_scheduler import ConstantLRSchedulerConfig
from flsim.privacy.common import PrivacySetting
from flsim.tests import utils
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
)
from omegaconf import OmegaConf
from opacus import PrivacyEngine, privacy_analysis


def calc_eps(sample_rate, noise_multiplier, steps, alphas, delta):
    rdp = privacy_analysis.compute_rdp(sample_rate, noise_multiplier, steps, alphas)
    eps, _ = privacy_analysis.get_privacy_spent(alphas, rdp, delta=delta)
    return eps


@pytest.fixture(scope="class")
def prepare_client_test(request):
    request.cls.num_batches = 2
    request.cls.batch_size = 10


@pytest.mark.usefixtures("prepare_client_test")
class ClientTestBase:
    def _fake_data(self, num_batches=None, batch_size=None):
        num_batches = num_batches or self.num_batches
        batch_size = batch_size or self.batch_size
        torch.manual_seed(0)
        dataset = [
            ([None] * batch_size, torch.rand(batch_size, 2)) for _ in range(num_batches)
        ]
        dataset = utils.DatasetFromList(dataset)
        return utils.DummyUserData(dataset, utils.SampleNet(utils.TwoFC()))

    def _get_client(
        self, data=None, store_models_and_optimizers=False, timeout_simulator=None
    ):
        data = data or self._fake_data()
        config = ClientConfig(
            store_models_and_optimizers=store_models_and_optimizers,
            lr_scheduler=ConstantLRSchedulerConfig(),
        )
        return Client(
            **OmegaConf.structured(config),
            dataset=data,
            timeout_simulator=timeout_simulator,
        )

    def _get_dp_client(
        self,
        data=None,
        noise_multiplier=1,
        clipping_value=1,
        store_models_and_optimizers=False,
    ):
        privacy_setting = PrivacySetting(
            noise_multiplier=noise_multiplier, clipping_value=clipping_value
        )
        config = DPClientConfig(
            store_models_and_optimizers=store_models_and_optimizers,
            privacy_setting=privacy_setting,
        )
        return DPClient(
            **OmegaConf.structured(config), dataset=(data or self._fake_data())
        )

    def _train(self, batches, model, optim):
        # basically re-write training logic
        model.fl_get_module().train()
        for batch in batches:
            optim.zero_grad()
            _batch = model.fl_create_training_batch(batch)
            loss = model.fl_forward(_batch).loss
            loss.backward()
            optim.step()

    def _run_client_eval_test(self, client):
        """
        Test client eval will turn on eval mode and turn back into train mode
        after evaluation loop is finished
        """

        class Net(utils.SampleNet):
            def get_eval_metrics(self, batch):
                assert (
                    self.sample_nn.training is False
                ), "Client should call eval after setting model.eval()"
                return self.sample_nn(batch)

        n_batches = 2
        input_dim = 2
        data = [torch.randn(input_dim, requires_grad=True) for _ in range(n_batches)]
        model = Net(nn.Linear(input_dim, 1))

        model.fl_get_module().train()
        client.eval(model=model, dataset=data)
        assert model.fl_get_module().training


class TestBaseClient(ClientTestBase):
    def test_storage(self):
        client = self._get_client(store_models_and_optimizers=True)
        model0 = utils.SampleNet(utils.TwoFC())
        model1 = utils.SampleNet(utils.TwoFC())
        delta1, weight1 = client.generate_local_update(model0)
        delta1 = deepcopy(delta1)

        delta2, weight2 = client.generate_local_update(model1)
        assertEqual(client.times_selected, 2)
        # model1 should be the first model stored
        assertAlmostEqual(weight1, client.weights[0])
        mismatched = utils.verify_models_equivalent_after_training(
            delta1, client.model_deltas[0]
        )
        assertEqual(mismatched, "", mismatched)
        # model2 should be second model stored
        assertAlmostEqual(weight2, client.weights[1])
        mismatched = utils.verify_models_equivalent_after_training(
            delta2, client.model_deltas[1]
        )
        assertEqual(mismatched, "", mismatched)

    def test_receive_through_channel(self):
        # expected channel effects,
        clnt = self._get_client()
        model = utils.SampleNet(utils.TwoFC())
        # check channel is pass through
        model2 = clnt.receive_through_channel(model)
        mismatched = utils.verify_models_equivalent_after_training(model2, model)
        assertEqual(mismatched, "", mismatched)

    def test_prepare_for_training(self):
        clnt = self._get_client()
        model = utils.SampleNet(utils.TwoFC())
        try:
            # should work
            model2, optim, optim_sch = clnt.prepare_for_training(model)
        except BaseException as e:
            assertTrue(False, e)
        mismatched = utils.verify_models_equivalent_after_training(model2, model)
        assertEqual(mismatched, "", mismatched)
        # expect correct type of optimizer
        assertIsInstance(optim, LocalOptimizerSGD)
        assertIsInstance(optim_sch, ConstantLRScheduler)

    def test_train(self):
        data = self._fake_data(num_batches=5, batch_size=10)
        clnt = self._get_client(data)
        model = utils.SampleNet(utils.TwoFC())
        model, optim, optim_sch = clnt.prepare_for_training(model)
        model2, optim2, _ = clnt.prepare_for_training(deepcopy(model))
        # value chekd in previous test
        try:
            # should work
            model, weight = clnt.train(model, optim, optim_sch, None)
        except BaseException as e:
            assertTrue(False, e)
        assertAlmostEqual(weight, 5 * 10)
        self._train(data, model2, optim2)
        mismatched = utils.verify_models_equivalent_after_training(model2, model)
        assertEqual(mismatched, "", mismatched)

    def test_generate_local_update(self):
        clnt = self._get_client()
        model = utils.SampleNet(utils.TwoFC())
        model.fl_get_module().fill_all(0.1)
        clnt.train = MagicMock(return_value=(model, 12.34))
        clnt.compute_delta = MagicMock(return_value=model)

        try:
            # should work
            delta, weight = clnt.generate_local_update(model)
        except BaseException as e:
            assertTrue(False, e)
        assertAlmostEqual(12.34, weight)
        mismatched = utils.verify_models_equivalent_after_training(model, model)
        assertEqual(mismatched, "", mismatched)
        mismatched = utils.verify_models_equivalent_after_training(delta, model)
        assertEqual(mismatched, "", mismatched)

    def test_fed_prox_sgd_equivalent(self):
        """
        Test FedProx under the following scenarios:

        FedProx == SGD iff
        1. FedProx with mu = 0 == SGD
        2. FedProx with mu = x == SGD(weight_decay=mu)

        FedProx != SGD if
        1. mu > 0 and SGD(weight_decay=0)
        """
        # scenario 1
        data = self._fake_data()
        prox_client = Client(
            dataset=data,
            **OmegaConf.structured(
                ClientConfig(
                    optimizer=LocalOptimizerFedProxConfig(mu=0),
                )
            ),
        )
        client = Client(
            dataset=data,
            **OmegaConf.structured(
                ClientConfig(
                    optimizer=LocalOptimizerSGDConfig(),
                )
            ),
        )

        init_model = utils.SampleNet(utils.TwoFC())
        delta, weight = client.generate_local_update(deepcopy(init_model))
        prox_delta, weight = prox_client.generate_local_update(deepcopy(init_model))
        mismatched = utils.verify_models_equivalent_after_training(
            prox_delta, delta, init_model
        )
        assertEqual(mismatched, "", mismatched)

        # scenario 2
        init_model = utils.SampleNet(utils.TwoFC())
        init_model.fl_get_module().fill_all(0.0)
        mu = 1.0
        prox_client = Client(
            dataset=data,
            **OmegaConf.structured(
                ClientConfig(
                    optimizer=LocalOptimizerFedProxConfig(mu=mu),
                )
            ),
        )
        client = Client(
            dataset=data,
            **OmegaConf.structured(
                ClientConfig(
                    optimizer=LocalOptimizerSGDConfig(weight_decay=mu),
                )
            ),
        )

        delta, _ = client.generate_local_update(deepcopy(init_model))
        prox_delta, _ = prox_client.generate_local_update(deepcopy(init_model))

        mismatched = utils.verify_models_equivalent_after_training(
            prox_delta, delta, init_model
        )
        assertEqual(mismatched, "", mismatched)

        # negative case
        # FedProx != SGD if mu > 0 and SGD has no weight decay
        init_model = utils.SampleNet(utils.TwoFC())
        init_model.fl_get_module().fill_all(0.0)
        mu = 1.0
        prox_client = Client(
            dataset=data,
            **OmegaConf.structured(
                ClientConfig(
                    optimizer=LocalOptimizerFedProxConfig(mu=mu),
                )
            ),
        )
        client = Client(
            dataset=data,
            **OmegaConf.structured(
                ClientConfig(
                    optimizer=LocalOptimizerSGDConfig(weight_decay=0),
                )
            ),
        )

        delta, _ = client.generate_local_update(deepcopy(init_model))
        prox_delta, _ = prox_client.generate_local_update(deepcopy(init_model))

        mismatched = utils.verify_models_equivalent_after_training(
            prox_delta, delta, init_model
        )
        assertNotEqual(mismatched, "", mismatched)

    def test_device_perf_generation(self):
        """
        test either client.device_perf is always generated,
        either using the TimeOutSimulator given in __init__
        or creates a NeverTimeOutSimulator( if none is provided.
        """
        data = self._fake_data(num_batches=5, batch_size=10)
        # pass gaussian timeout into client
        cfg = GaussianTimeOutSimulatorConfig(
            timeout_wall_per_round=1.0,
            fl_stopping_time=1.0,
            duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                training_duration_mean=1.0, training_duration_sd=0.0
            ),
        )
        gaussian_timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(cfg)
        )
        clnt_gaussian_timeout = self._get_client(
            data, timeout_simulator=gaussian_timeout_simulator
        )
        assertEqual(clnt_gaussian_timeout.per_example_training_time, 1.0)
        # pass never timeout to clients
        clnt_never_timeout = self._get_client(
            data,
            timeout_simulator=NeverTimeOutSimulator(
                **OmegaConf.structured(NeverTimeOutSimulatorConfig())
            ),
        )
        assertEqual(clnt_never_timeout.per_example_training_time, 0.0)
        # default created never timeout within clients
        clnt_default = self._get_client(data)
        assertEqual(clnt_default.per_example_training_time, 0.0)

    def test_total_training_time(self):
        """
        total training time for Gaussian with mean 1.0 and std 0.0
        equals to min(number of client examples, timeout_wall_per_round)
        """
        data = self._fake_data(num_batches=5, batch_size=10)
        # pass gaussian timeout into client with a big timeout wall
        cfg_big_wall = GaussianTimeOutSimulatorConfig(
            timeout_wall_per_round=99999.0,
            fl_stopping_time=1.0,
            duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                training_duration_mean=1.0, training_duration_sd=0.0
            ),
        )
        gaussian_timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(cfg_big_wall)
        )
        clnt_gaussian_timeout = self._get_client(
            data, timeout_simulator=gaussian_timeout_simulator
        )
        num_examples = 5 * 10
        assertEqual(clnt_gaussian_timeout.get_total_training_time(), num_examples)
        # pass gaussian timeout into client with a small timeout wall
        timeout_wall = 25.0
        cfg_small_wall = GaussianTimeOutSimulatorConfig(
            timeout_wall_per_round=timeout_wall,
            fl_stopping_time=1.0,
            duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                training_duration_mean=1.0, training_duration_sd=0.0
            ),
        )
        gaussian_timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(cfg_small_wall)
        )
        clnt_gaussian_timeout = self._get_client(
            data, timeout_simulator=gaussian_timeout_simulator
        )
        assertEqual(clnt_gaussian_timeout.get_total_training_time(), timeout_wall)
        # pass never timeout to clients
        clnt_never_timeout = self._get_client(
            data,
            timeout_simulator=NeverTimeOutSimulator(
                **OmegaConf.structured(NeverTimeOutSimulatorConfig())
            ),
        )
        assertEqual(clnt_never_timeout.get_total_training_time(), 0.0)

    def test_partial_training(self):
        """
        producing two training instance expected same training results:
        1. client with n (even number) batches and a time out wall just
        enough for training half of the batches
        2. client with n/2 batches and no time out wall

        check model1 == model2
        """
        n_batches = 2
        bs = 10
        data = self._fake_data(num_batches=n_batches, batch_size=bs)
        # only feasible to process 3 batches, client sends back partial results
        expected_processed_samples = n_batches * bs / 2
        cfg = GaussianTimeOutSimulatorConfig(
            timeout_wall_per_round=expected_processed_samples,
            fl_stopping_time=1.0,
            duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                training_duration_mean=1.0, training_duration_sd=0.0
            ),
        )
        gaussian_timeout_simulator = GaussianTimeOutSimulator(
            **OmegaConf.structured(cfg)
        )
        clnt_gaussian_timeout = self._get_client(
            data, timeout_simulator=gaussian_timeout_simulator
        )
        torch.manual_seed(0)
        model_init = utils.SampleNet(utils.TwoFC())
        model1, optim1, optim_sch1 = clnt_gaussian_timeout.prepare_for_training(
            deepcopy(model_init)
        )
        torch.manual_seed(0)
        partial_model, partial_weight = clnt_gaussian_timeout.train(
            model1, optim1, optim_sch1, None
        )
        assertEqual(partial_weight, expected_processed_samples)

        # no timeout, but client only has half of the data
        n_batches = int(n_batches / 2)
        data = self._fake_data(num_batches=n_batches, batch_size=bs)
        clnt = self._get_client(data)
        model2, optim2, optim_sch2 = clnt.prepare_for_training(deepcopy(model_init))
        torch.manual_seed(0)
        full_model, full_weight = clnt.train(model2, optim2, optim_sch2, None)
        assertEqual(full_weight, expected_processed_samples)

        mismatched = utils.verify_models_equivalent_after_training(
            full_model, partial_model
        )
        assertEqual(mismatched, "", mismatched)

    def test_logging_level(self):
        clnt = self._get_client()
        assertTrue(utils.check_inherit_logging_level(clnt, 50))
        assertTrue(utils.check_inherit_logging_level(clnt, 10))

    def test_base_client_eval(self):
        client = self._get_client()
        self._run_client_eval_test(client)


class TestDPClient(ClientTestBase):
    def test_privacy_engine_properly_initialized(self):
        data = self._fake_data(num_batches=11, batch_size=3)
        clnt = self._get_dp_client(data, noise_multiplier=0.1, clipping_value=2.0)
        model = utils.SampleNet(utils.TwoFC())
        model, optim, optim_sch = clnt.prepare_for_training(model)
        p_eng = optim.privacy_engine
        assertEqual(p_eng.noise_multiplier, 0.1)
        assertEqual(p_eng.max_grad_norm, 2.0)
        assertEqual(p_eng.sample_rate, 1.0 / 11)

    def test_privacy_turned_off(self):
        data = self._fake_data(num_batches=11, batch_size=3)
        # clipping value of inf means privacy is off no matter what the noise multiplier
        clnt = self._get_dp_client(
            data, noise_multiplier=0.0, clipping_value=float("inf")
        )
        assertFalse(clnt.privacy_on)
        clnt = self._get_dp_client(
            data, noise_multiplier=1.0, clipping_value=float("inf")
        )
        assertFalse(clnt.privacy_on)
        # negative noise multiplier should turn of privacy engine
        clnt = self._get_dp_client(data, noise_multiplier=-1.0, clipping_value=0.1)
        assertFalse(clnt.privacy_on)

    def test_prepare_for_training(self):
        clnt = self._get_dp_client()
        model = utils.SampleNet(utils.TwoFC())
        model2, optim, optim_sch = clnt.prepare_for_training(model)
        mismatched = utils.verify_models_equivalent_after_training(model2, model)
        assertEqual(mismatched, "")
        # expect correct type of optimizer
        assertIsInstance(optim, LocalOptimizerSGD)
        assertIsInstance(optim_sch, ConstantLRScheduler)
        # expect privacy engine attached
        p_eng = optim.privacy_engine
        assertTrue(hasattr(p_eng, "optimizer"))

    def test_storage(self):
        client = self._get_dp_client(store_models_and_optimizers=True)
        model0 = utils.SampleNet(utils.TwoFC())
        delta, weight1 = client.generate_local_update(model0)

        assertEqual(client.times_selected, 1)
        # test existance of privacy_engine
        # model1 should be the first model stored
        optim = client.optimizers[0]
        assertTrue(hasattr(optim, "privacy_engine"))
        assertIsInstance(optim.privacy_engine, PrivacyEngine)

    def test_no_noise_no_clip(self):
        data = self._fake_data(3, 4)
        model = utils.SampleNet(utils.TwoFC())
        private_model = deepcopy(model)

        clnt = self._get_client(data)
        delta, weight = clnt.generate_local_update(model)
        # set noise to 0 and clipping to a large number
        private_clnt = self._get_dp_client(
            data, noise_multiplier=0, clipping_value=1000
        )

        private_delta, private_weight = private_clnt.generate_local_update(
            private_model
        )
        mismatched = utils.verify_models_equivalent_after_training(model, private_model)
        mismatched_delta = utils.verify_models_equivalent_after_training(
            delta, private_delta
        )
        assertAlmostEqual(weight, private_weight)
        assertEqual(mismatched, "", mismatched)
        assertEqual(mismatched_delta, "", mismatched_delta)

    def test_only_clip(self):
        data = self._fake_data(4, 4)
        model = utils.SampleNet(utils.TwoFC())
        private_model = deepcopy(model)

        clnt = self._get_client(data)
        delta, weight = clnt.generate_local_update(model)
        private_clnt = self._get_dp_client(
            data, noise_multiplier=0, clipping_value=0.01
        )

        private_delta, private_weight = private_clnt.generate_local_update(
            private_model
        )
        mismatched = utils.verify_models_equivalent_after_training(delta, private_delta)
        assertAlmostEqual(weight, private_weight)
        assertNotEqual(mismatched, "")

    def test_noise_and_clip(self):
        data = self._fake_data(4, 4)
        model = utils.SampleNet(utils.TwoFC())
        private_model = deepcopy(model)

        clnt = self._get_client(data)
        delta, weight = clnt.generate_local_update(model)
        private_clnt = self._get_dp_client(
            data, noise_multiplier=1, clipping_value=0.01
        )
        private_delta, private_weight = private_clnt.generate_local_update(
            private_model
        )
        mismatched_delta = utils.verify_models_equivalent_after_training(
            delta, private_delta
        )
        assertAlmostEqual(weight, private_weight)
        assertNotEqual(mismatched_delta, "")

    def test_epsilon(self):
        noise_multiplier = 1.5
        clipping_value = 0.01
        num_batches = 5
        batch_size = 6
        data = self._fake_data(num_batches, batch_size)
        model = utils.SampleNet(utils.TwoFC())
        clnt = self._get_dp_client(data, noise_multiplier, clipping_value, True)
        model, weight = clnt.generate_local_update(model)
        privacy_engine = clnt.optimizers[0].privacy_engine
        alphas = privacy_engine.alphas
        delta = privacy_engine.target_delta

        eps_from_script = calc_eps(
            1.0 / num_batches, noise_multiplier, num_batches, alphas, delta
        )
        eps_from_client, _ = privacy_engine.get_privacy_spent()
        assertAlmostEqual(eps_from_script, eps_from_client)

    def test_dp_client_eval(self):
        dp_client = self._get_dp_client()
        self._run_client_eval_test(dp_client)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import torch
from flsim.common.pytest_helper import assertEqual, assertNotEqual
from flsim.secure_aggregation.secure_aggregator import FixedPointConfig
from flsim.servers.aggregator import AggregationType
from flsim.servers.sync_secagg_servers import SyncSecAggServerConfig
from flsim.servers.sync_servers import SyncServerConfig
from flsim.tests.utils import (
    FakeMetricReporter,
)
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.sync_trainer_test_utils import (
    create_sync_trainer,
)
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset


class TestSecureAggregationIntegration:
    def _load_data(self, num_users: int = 26):
        """
        Loads the data, which is a Dummy alphabet for N (`num_users`) users,
        each with 1 sample.
        """
        shard_size = 1
        local_batch_size = 1
        dummy_dataset = DummyAlphabetDataset(num_rows=num_users)
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, DummyAlphabetFLModel()
        )
        assertEqual(data_loader.num_total_users, num_users / shard_size)
        assertEqual(data_loader.num_total_users, data_provider.num_users())
        return data_provider, data_loader.train_batch_size

    def _train_fl_model(
        self,
        sec_agg_enable: bool = False,
        fixedpoint=None,
        num_users: int = 26,
    ):
        """
        Trains an FL model, with or without Secure Aggregation
        """
        # create dummy FL model on alphabet
        global_fl_model = DummyAlphabetFLModel()
        data_provider, _ = self._load_data(num_users)
        world_size = 1
        sync_trainer = create_sync_trainer(
            model=global_fl_model,
            local_lr=0.1,
            users_per_round=num_users,
            epochs=1,
            user_epochs_per_round=1,
            do_eval=True,
            server_config=SyncSecAggServerConfig(
                aggregation_type=AggregationType.AVERAGE, fixedpoint=fixedpoint
            )
            if sec_agg_enable
            else SyncServerConfig(
                aggregation_type=AggregationType.AVERAGE,
            ),
        )
        metrics_reporter = FakeMetricReporter()
        global_fl_model, _eval_metric = sync_trainer.train(
            data_provider,
            metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        return global_fl_model

    def test_secagg_not_equivalent_no_secagg(self) -> None:
        """
        Tests training with secure aggregation will produce a different
        model than training without secure aggregation
        """
        # First, call SyncTrainer with SecureRoundReducer
        fixedpoint = FixedPointConfig(num_bytes=1, scaling_factor=1000)
        torch.manual_seed(1)
        fl_model_with_secure_round_reducer = self._train_fl_model(
            sec_agg_enable=True,
            fixedpoint=fixedpoint,
        )
        # Next, call SyncTrainer (with RoundReducer)
        torch.manual_seed(1)
        fl_model_with_round_reducer = self._train_fl_model()
        assertNotEqual(
            FLModelParamUtils.get_mismatched_param(
                [
                    fl_model_with_round_reducer.fl_get_module(),
                    fl_model_with_secure_round_reducer.fl_get_module(),
                ],
                1e-6,
            ),
            "",
        )

    def test_secagg_not_equivalent_no_secagg_large_range(self) -> None:
        """
        Tests training with secure aggregation will produce a different
        model than training without secure aggregation, even when the
        range of fixedpoint number is very large (and scaling factro is 1).

        The reason that we get a different model is because, even in a big
        fixedpoint range, we still do rounding when we convert, E.g., 127.1
        (float) becomes 127 (in fixedpoint), no matter how big the range is.
        """
        # First, call SyncTrainer with SecureRoundReducer
        fixedpoint = FixedPointConfig(num_bytes=8, scaling_factor=1)
        torch.manual_seed(1)
        fl_model_with_secure_round_reducer = self._train_fl_model(
            sec_agg_enable=True,
            fixedpoint=fixedpoint,
        )
        # Next, call SyncTrainer (with RoundReducer)
        torch.manual_seed(1)
        fl_model_with_round_reducer = self._train_fl_model()
        assertNotEqual(
            FLModelParamUtils.get_mismatched_param(
                [
                    fl_model_with_round_reducer.fl_get_module(),
                    fl_model_with_secure_round_reducer.fl_get_module(),
                ],
                1e-6,
            ),
            "",
        )



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import abc

import torch
from flsim.interfaces.model import IFLModel


def FloatTensor(cuda_enabled: bool, *args):
    if cuda_enabled:
        return torch.cuda.FloatTensor(*args)
    else:
        return torch.FloatTensor(*args)


def tensor(data, dtype, cuda_enabled: bool):
    return torch.tensor(data, dtype=dtype, device=device(cuda_enabled))


def device(cuda_enabled: bool):
    return "cuda:{}".format(torch.cuda.current_device()) if cuda_enabled else "cpu"


class ICudaStateManager(abc.ABC):
    """Sends model from CPU-->GPU, and from GPU-->CPU,
    if required, at 3 different times:
    a) When trainer is initialized
    b) Before training or eval is done
    c) After training or eval is done
    Centralizes all CPU-->GPU/GPU-->CPU moves
    """

    @abc.abstractmethod
    def on_trainer_init(self, model: IFLModel):
        pass

    @abc.abstractmethod
    def before_train_or_eval(self, model: IFLModel):
        pass

    @abc.abstractmethod
    def after_train_or_eval(self, model: IFLModel):
        pass


class NoopCudaStateManager(ICudaStateManager):
    def __init__(self):
        pass

    def on_trainer_init(self, model: IFLModel):
        pass

    def before_train_or_eval(self, model: IFLModel):
        pass

    def after_train_or_eval(self, model: IFLModel):
        pass


class CudaTransferMinimizer(ICudaStateManager):
    """Minimize CPU<-->GPU memory bandwidth,
    at the cost of increasing GPU memory consumption
    Model is moved to GPU right when trainer is initialized.
    All model copies stay on the GPU.
    E.g: when Sync trainer creates clients, all of them
    get models that are already in GPU
    GPU<-->GPUMemory bandwidth is really high, 2 orders of magnitude
    larger than CPUMemory<-->GPUMemory bandwidth.
    THIS SHOULD BE THE DEFAULT UNLESS RUNNING OUT OF GPU MEMORY
    """

    def __init__(self, cuda_enabled: bool):
        self._cuda_enabled = cuda_enabled

    def on_trainer_init(self, model: IFLModel):
        """When trainer is initialized, we move the model to GPU
        Any furter copies of the model will stay on GPU
        """
        if self._cuda_enabled:
            model.fl_cuda()

    def before_train_or_eval(self, model: IFLModel):
        pass

    def after_train_or_eval(self, model: IFLModel):
        pass


class GPUMemoryMinimizer(ICudaStateManager):
    """Minimize GPU memory at the cost of increasing
    CPU-->GPU and GPU-->CPU memory bandwidth consumption
    Model is moved to GPU right before train/eval is called,
    and moved out of GPU right after.
    All operations other than training/eval happen on CPU.
    E.g: global model aggregation happens on CPU

    DONT USE THIS UNLESS YOU ARE RUNNING OUT OF GPU MEMORY
    """

    def __init__(self, cuda_enabled: bool):
        self._cuda_enabled = cuda_enabled

    def on_trainer_init(self, model: IFLModel):
        if self._cuda_enabled:
            model.fl_get_module().to("cpu")

    def before_train_or_eval(self, model: IFLModel):
        if self._cuda_enabled:
            model.fl_cuda()

    def after_train_or_eval(self, model: IFLModel):
        if self._cuda_enabled:
            model.fl_get_module().to("cpu")


# default manager, does nothing
DEFAULT_CUDA_MANAGER: ICudaStateManager = NoopCudaStateManager()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.


from dataclasses import dataclass, field
from typing import List, NamedTuple, Optional

import numpy as np


@dataclass
class PrivacySetting:
    """
    Contains setting related to Differential Privacy
    """

    alphas: List[float] = field(
        default_factory=lambda: np.arange(1.1, 100, 0.1).tolist()
    )  # Renyi privacy alpha range
    noise_multiplier: float = 0.0  # Normalized Noise Variance
    clipping_value: float = float("inf")
    target_delta: float = 1e-5  # Maximum delta for (epsilon, delta) privacy
    noise_seed: Optional[int] = None  # [debug] Seed of the noise generation function
    secure_rng: bool = False


class PrivacyBudget(NamedTuple):
    """
    Encapsulates a privacy budget as (epsilon, delta)
    """

    epsilon: float = float("inf")
    alpha: float = float(-1)
    delta: float = float("inf")

    def __str__(self):
        return f"eps = {self.epsilon}, delta = {self.delta}, alpha = {self.alpha}"



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
import copy
from dataclasses import dataclass
from typing import Any, Optional

from flsim.interfaces.batch_metrics import IFLBatchMetrics
from flsim.interfaces.model import IFLModel
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from flsim.utils.fl.common import FLModelParamUtils
from omegaconf import MISSING
from torch.optim.optimizer import Optimizer


class OptimizerScheduler(abc.ABC):
    """
    base class for local LR scheduler, enable the laerning rate
    of local optimizers for individual users during local training
    """

    def __init__(
        self,
        *,
        optimizer: Optimizer,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=OptimizerSchedulerConfig,
            **kwargs,
        )
        self.optimizer = optimizer

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    @abc.abstractmethod
    def step(
        self,
        batch_metric: Optional[IFLBatchMetrics] = None,
        model: Optional[IFLModel] = None,
        data: Optional[Any] = None,
        epoch: Optional[int] = None,
    ):
        """
        interface for updating learning rate. Some learning rate scheduling methods
        rely produces multiple trial forward passes internally, e.g, line search
        methods, hence model is required in the interface.
        """
        pass

    def get_lr(self):
        lrs = [param_group["lr"] for param_group in self.optimizer.param_groups]
        return lrs


class ConstantLRScheduler(OptimizerScheduler):
    def __init__(
        self,
        *,
        optimizer: Optimizer,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ConstantLRSchedulerConfig,
            **kwargs,
        )

        super().__init__(optimizer=optimizer, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def step(
        self,
        batch_metric: Optional[IFLBatchMetrics] = None,
        model: Optional[IFLModel] = None,
        data: Optional[Any] = None,
        epoch: Optional[int] = None,
    ):
        pass

    @property
    def lr(self):
        return self.cfg.base_lr


class LRBatchSizeNormalizer(OptimizerScheduler):
    """
    normalized the LR by number of examples in the batch
    """

    def __init__(
        self,
        *,
        optimizer: Optimizer,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=LRBatchSizeNormalizerSchedulerConfig,
            **kwargs,
        )

        super().__init__(optimizer=optimizer, **kwargs)

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def step(
        self,
        batch_metric: Optional[IFLBatchMetrics],
        model: Optional[IFLModel] = None,
        data: Optional[Any] = None,
        epoch: Optional[int] = None,
    ):
        assert (
            batch_metric is not None
        ), "`batch_metric` param cannot be None for LRBatchSizeNormalizer"
        lr_normalizer = self._get_lr_normalizer(batch_metric)
        for param_group in self.optimizer.param_groups:
            # pyre-fixme[16]: `LRBatchSizeNormalizer` has no attribute `cfg`.
            param_group["lr"] = self.cfg.base_lr * lr_normalizer

    def _get_lr_normalizer(self, batch_metric: IFLBatchMetrics):
        # pyre-fixme[16]: `LRBatchSizeNormalizer` has no attribute `cfg`.
        return batch_metric.num_examples / self.cfg.local_lr_normalizer


class ArmijoLineSearch(OptimizerScheduler):
    """
    Classical Armijo line-search for step-size selection in optimization.
    Recent work suggests that it might also be used in stochastic over-parametrized
    setting, cf.
    "Painless Stochastic Gradient: Interpolation, Line-Search, and Convergence Rates"
    """

    def __init__(
        self,
        *,
        optimizer: Optimizer,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ArmijoLineSearchSchedulerConfig,
            **kwargs,
        )

        super().__init__(optimizer=optimizer, **kwargs)
        assert (
            0
            # pyre-fixme[16]: `ArmijoLineSearch` has no attribute `cfg`.
            < self.cfg.shrinking_factor
            <= 1.0
        ), "shrinking_factor must be between 0 and 1.0"
        assert 0 < self.cfg.c <= 1.0, "constant c must be between 0 and 1.0"
        assert (
            self.cfg.max_iter > 0
        ), "number of line-search iterations must be a non-negative integer"

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def step(
        self,
        batch_metric: Optional[IFLBatchMetrics],
        model: Optional[IFLModel],
        data: Optional[Any] = None,
        epoch: Optional[int] = None,
    ):
        assert (
            batch_metric is not None
        ), "`batch_metric` param cannot be None for ArmijoLineSearch"
        assert model is not None, "`model` param cannot be None for ArmijoLineSearch"

        state_dict = copy.deepcopy(
            FLModelParamUtils.get_state_dict(
                model.fl_get_module(), only_federated_params=False
            )
        )
        grad_norm_before_update = FLModelParamUtils.get_gradient_l2_norm_raw(
            model.fl_get_module()
        )
        loss_before_update = batch_metric.loss.item()
        # pyre-fixme[16]: `ArmijoLineSearch` has no attribute `cfg`.
        if self.cfg.reset:
            self._reset_lr()

        for _ in range(self.cfg.max_iter):
            FLModelParamUtils.load_state_dict(
                model.fl_get_module(), state_dict, only_federated_params=False
            )
            proposed_lr = self.get_lr()
            assert (
                len(proposed_lr) == 1
            ), "Armijo line-search only works with single param_group"
            # pyre-ignore[20]
            self.optimizer.step()
            # DO NOT compute backprop after forward here, only the forward is
            # required for step-size selection, use existent gradient direction
            new_batch_metrics = model.fl_forward(data)
            # loss if we use the proposed LR
            new_loss = new_batch_metrics.loss.item()
            if (
                float(new_loss)
                <= loss_before_update
                - self.cfg.c * proposed_lr[0] * grad_norm_before_update ** 2
            ):
                # satisfy sufficient descent, accept proposed_lr
                # and terminate line search
                break
            # reduce lr
            self._shrink_lr()

        # recover model state before the line search scatching, do the actual
        # optimizer.step() outside of the scheduler
        FLModelParamUtils.load_state_dict(
            model.fl_get_module(), state_dict, only_federated_params=False
        )

    def _shrink_lr(self):
        for param_group in self.optimizer.param_groups:
            param_group["lr"] *= self.cfg.shrinking_factor

    def _reset_lr(self):
        # reset LR back to base lr, use for resetting LR across training batches
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = self.cfg.base_lr


@dataclass
class OptimizerSchedulerConfig:
    _target_: str = MISSING
    _recursive_: bool = False
    base_lr: float = 0.001


@dataclass
class ConstantLRSchedulerConfig(OptimizerSchedulerConfig):
    _target_: str = fullclassname(ConstantLRScheduler)


@dataclass
class LRBatchSizeNormalizerSchedulerConfig(OptimizerSchedulerConfig):
    _target_: str = fullclassname(LRBatchSizeNormalizer)
    local_lr_normalizer: int = 1


@dataclass
class ArmijoLineSearchSchedulerConfig(OptimizerSchedulerConfig):
    _target_: str = fullclassname(ArmijoLineSearch)
    # between (0, 1), algorithm parameter, no need to sweep usually
    shrinking_factor: float = 0.5
    # between (0, 1), algorithm parameter, no need to sweep usually
    c: float = 0.5
    # whether to reset the learning rate to base_lr in between steps
    # if False, line search for next optimizer.step() will continue
    # from the step-size found in the previous step
    reset: bool = False
    # maximum number of line-search iterations
    max_iter: int = 5



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import random
from dataclasses import dataclass
from typing import List

import numpy as np
import pytest
import torch
from flsim.common.pytest_helper import assertTrue, assertEqual, assertEmpty
from flsim.interfaces.model import IFLModel
from flsim.optimizers.async_aggregators import (
    AsyncAggregationType,
    AsyncAggregatorConfig,
    FedAvgWithLRAsyncAggregatorConfig,
    FedAvgWithLRWithMomentumAsyncAggregatorConfig,
    FedAvgWithLRHybridAggregatorConfig,
)
from flsim.tests.utils import MockQuadratic1DFL, Quadratic1D, SampleNet, TwoFC
from flsim.tests.utils import (
    verify_models_equivalent_after_training,
)
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate


@dataclass
class MockClientModel:
    delta: SampleNet
    after_train: SampleNet
    weight: float


class TestAsyncAggregator:
    def _test_one_step(
        self,
        param_after_local_training: float,
        param_after_global_training: float,
        weight: float,
        config: AsyncAggregatorConfig,
    ):
        """
        Test async aggregator by:
        1. Create AsyncAggregator of given type (fed_buff_aggregation or fed_async_aggregation)
        2. Set local param after training to param_after_local_training
        3. Set global param (simulate update from another user) to
           param_after_global_training
        4. Verify that async_optimizer.step(weight=weight) sets global param to correct value
        """
        # initial parameters: x=1.0, y=1.0 (x=param, y=constant)
        init_val = 1
        global_model = MockQuadratic1DFL(Quadratic1D())
        async_aggregator = instantiate(config, global_model=global_model)
        local_model = copy.deepcopy(global_model)
        delta = copy.deepcopy(global_model)
        # local training, x becomes param_after_local_training
        local_model.fl_get_module().x.data = torch.Tensor([param_after_local_training])
        # global update from another user, x becomes param_after_global_training
        global_model.fl_get_module().x.data = torch.Tensor(
            [param_after_global_training]
        )
        # client delta is init_val - param_after_local_training
        delta.fl_get_module().x.data = torch.Tensor(
            [init_val - param_after_local_training]
        )

        async_aggregator.on_client_training_end(
            client_delta=delta,
            final_local_model=local_model,
            weight=weight,
        )
        if config.aggregation_type == AsyncAggregationType.fed_buff_aggregation:
            # model delta = final_local_model - local_model_deltaing = param_after_local_training  - 1
            # global_model = param_after_global_training + weight*(param_after_local_training - 1)
            global_model_expected = param_after_global_training + weight * (
                param_after_local_training - init_val
            )
        else:
            # global_model = (1- weight)*param_after_global_training + weight*param_after_local_training
            global_model_expected = (
                1 - weight
            ) * param_after_global_training + weight * param_after_local_training

        assertTrue(
            torch.allclose(
                global_model.fl_get_module().x.data,
                torch.Tensor([global_model_expected]),
                atol=1e-7,
            )
        )

    def test_fed_buff_aggregation(self):
        """
        Test that AsyncAggregator with 'fed_buff_aggregation' works as expected
        1. Create global_model, some_param=1
        2. Copy global_model into local_model
        3. Simulate local training by local_model changing some_param to a (some_param=a, delta=a-1)
        4. Simulate global model updata by other user, some_param being set to b (some_param=b)
        5. Call AsyncOptimizer.step(). global_model.some_param should be b + (a-1) (some_param == b + a -1)
        """

        num_random_values = 5
        num_random_weights = 5
        for config in [
            FedAvgWithLRAsyncAggregatorConfig(
                aggregation_type=AsyncAggregationType.fed_buff_aggregation, lr=1.0
            ),
            FedAvgWithLRHybridAggregatorConfig(
                aggregation_type=AsyncAggregationType.fed_buff_aggregation,
                lr=1.0,
                buffer_size=1,
            ),
        ]:
            for _ in range(num_random_values):
                for _ in range(num_random_weights):
                    self._test_one_step(
                        param_after_local_training=np.random.random_sample(),
                        param_after_global_training=np.random.random_sample(),
                        weight=np.random.random_sample(),
                        config=config,
                    )

    def test_fed_async_aggregation(self):
        """
        Test that AsyncAggregator with 'fed_async_aggregation' works as expected
        1. Create global_model, some_param=1
        2. Copy global_model into local_model
        3. Simulate local training by local_model changing some_param to a (some_param=a)
        4. Simulate global model updata by other user, some_param being set to b (some_param=b)
        5. Call AsyncOptimizer.step(weight=1). global_model.some_param should be
            w*a + (1-w)*b (some_param == w*a + (1-w)*b)
        """
        num_random_values = 5
        num_random_weights = 5
        for _ in range(num_random_values):
            for _ in range(num_random_weights):
                self._test_one_step(
                    param_after_local_training=np.random.random_sample(),
                    param_after_global_training=np.random.random_sample(),
                    weight=np.random.random_sample(),
                    config=FedAvgWithLRAsyncAggregatorConfig(
                        aggregation_type=AsyncAggregationType.fed_async_aggregation,
                        lr=1.0,
                    ),
                )

    def _create_n_clients(self, num_clients):
        return [
            MockClientModel(
                delta=SampleNet(TwoFC()),
                after_train=SampleNet(TwoFC()),
                weight=np.random.random_sample(),
            )
            for _ in range(num_clients)
        ]

    def _symmetry_test(self, num_users, hybrid_config):

        hybrid_global_model_1 = SampleNet(TwoFC())
        hybrid_global_model_2 = copy.deepcopy(hybrid_global_model_1)

        hybrid_aggregator_1 = instantiate(
            hybrid_config, global_model=hybrid_global_model_1
        )

        hybrid_aggregator_2 = instantiate(
            hybrid_config, global_model=hybrid_global_model_2
        )

        client_models = self._create_n_clients(num_users)

        for client_model in client_models:
            hybrid_aggregator_1.zero_grad()
            hybrid_aggregator_1.on_client_training_end(
                client_model.delta,
                client_model.after_train,
                weight=client_model.weight,
            )

        random.shuffle(client_models)
        for client_model in client_models:
            hybrid_aggregator_2.zero_grad()
            hybrid_aggregator_2.on_client_training_end(
                client_model.delta,
                client_model.after_train,
                weight=client_model.weight,
            )

        return FLModelParamUtils.get_mismatched_param(
            models=[
                hybrid_global_model_1.fl_get_module(),
                hybrid_global_model_2.fl_get_module(),
            ],
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )

    def _equivalence_test(self, num_users, hybrid_config, async_config):
        async_global_model = SampleNet(TwoFC())
        hybrid_global_model = copy.deepcopy(async_global_model)

        async_aggregator = instantiate(async_config, global_model=async_global_model)

        hybrid_aggregator = instantiate(hybrid_config, global_model=hybrid_global_model)

        client_models = self._create_n_clients(num_users)

        for client_model in client_models:
            async_aggregator.zero_grad()
            async_aggregator.on_client_training_end(
                client_model.delta,
                client_model.after_train,
                weight=client_model.weight,
            )

        for client_model in client_models:
            hybrid_aggregator.zero_grad()
            hybrid_aggregator.on_client_training_end(
                client_model.delta,
                client_model.after_train,
                weight=client_model.weight,
            )

        return FLModelParamUtils.get_mismatched_param(
            models=[
                async_global_model.fl_get_module(),
                hybrid_global_model.fl_get_module(),
            ],
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )

    def test_hybrid_async_symmetry(self):
        """
        Test for symmetry:
        To satisfy symmetry, a hybrid async aggregation algorithm should be invariant to the order of user updates
        f(userA, userB) = f(userB, userA) where f() is aggregation mechanism

        1. Create async and hybrid aggregators with same global model
        2. Create a list of N clients
        3. Run hybrid_aggregator
        4. Shuffle client list
        5. Run async_aggregator
        6. Both should reach the same final global model
        """
        num_users = 10
        global_lr = 1.0
        hybrid_config = FedAvgWithLRHybridAggregatorConfig(lr=global_lr, buffer_size=1)

        error_msg = self._symmetry_test(
            num_users=num_users, hybrid_config=hybrid_config
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_hybrid_async_equivalence(self):
        """
        To satisfy equivalence,
        1. Assume both mechanisms have the same starting point
        2. Denote N = number of users
        3. Assume buffer_size is a factor of N
        4. Pure async and hybrid-async would reach the same final global model

        For simplicity, we assume buffer_size = N
        """
        num_users = 10
        global_lr = 1.00

        async_config = FedAvgWithLRAsyncAggregatorConfig(lr=global_lr)
        hybrid_config = FedAvgWithLRHybridAggregatorConfig(lr=global_lr, buffer_size=10)

        error_msg = self._equivalence_test(
            num_users=num_users, hybrid_config=hybrid_config, async_config=async_config
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_global_update(self):
        """
        Test the aggregator only updates global model if
        threshold is reached
        """
        num_epochs = 5
        for _ in range(num_epochs):
            num_total_users = np.random.randint(1, 20)
            buffer_size = np.random.randint(1, num_total_users + 1)

            hybrid_config = FedAvgWithLRHybridAggregatorConfig(
                lr=1.0, buffer_size=buffer_size
            )
            global_model = SampleNet(TwoFC())
            hybrid_aggregator = instantiate(hybrid_config, global_model=global_model)
            client_models = self._create_n_clients(num_total_users)
            for client_num, client in enumerate(client_models):
                is_global_model_updated = hybrid_aggregator.on_client_training_end(
                    client.delta, client.after_train, weight=1
                )
                # client_num is 0th index hence we need the + 1
                should_update_global_model = (client_num + 1) % buffer_size == 0
                assertEqual(is_global_model_updated, should_update_global_model)

    def train_async_with_zero_weight(
        self,
        initial_model: IFLModel,
        client_models: List[MockClientModel],
        num_epochs: int,
        num_total_users: int,
        momentum: float,
        train_with_zero_weight_in_middle: bool,
    ) -> IFLModel:
        """'Train' initial model by applying randomly generated client model updates to
        it, by repeatedly calling aggregator.on_client_training_end
        We do it thrice:
        a) Train for num_epochs/2
        b) If train_with_zero_weight_in_middle, train for num_epochs with zero weight
        c) Train for num_epochs/2
        Return final model
        """
        assert num_epochs % 2 == 0, "Training must be over even number of epochs"
        # config = AsyncAggregatorFedSGDConfig(lr=1.0, momentum=momentum)
        config = FedAvgWithLRWithMomentumAsyncAggregatorConfig(
            lr=1.0, momentum=momentum
        )
        aggregator = instantiate(config, global_model=initial_model)
        half_epochs = int(num_epochs / 2)

        def print_debug(prefix: str):
            for key, value in aggregator.optimizer.state.items():
                print(f"{prefix}: {key}:{value}")
                break

        for _ in range(half_epochs):
            for client in client_models:
                aggregator.on_client_training_end(
                    client.delta, client.after_train, weight=1
                )
        print_debug("After first loop")
        if train_with_zero_weight_in_middle:
            # training with zero weight should change neither the model, nor
            # the velocity computation inside the optimizer
            for _ in range(half_epochs):
                for client in client_models:
                    aggregator.on_client_training_end(
                        client.delta, client.after_train, weight=0
                    )
        print_debug("After second loop")
        for _ in range(half_epochs):
            for client in client_models:
                aggregator.on_client_training_end(
                    client.delta, client.after_train, weight=1
                )
        print_debug("After third loop")
        return aggregator.global_model

    @pytest.mark.parametrize(
        "num_total_users,num_epochs, momentum",
        [(1, 2, 0.5), (10, 10, 0.5), (10, 10, 0)],
    )
    def test_momentum_implementation_zero_weight(
        self, num_total_users, num_epochs, momentum
    ):
        """In FedAsyncAggregatorWithMomentum.on_client_training_end, when weight=0,
        neither velocity nor model should be updated
        We test this by comparing two training runs:
        RUN 1
            a) Running a few FL rounds in FedSyncAggregatorWithMomentum
            b) Calling on_client_training_end with weight=0
            c) Running some more FL rounds with FedSyncAggregatorWithMomentum
        RUN 2
            Same as RUN 1, except no (b)
        RUN 1 and RUN 2 should produce the same model
        """

        # function starts here
        initial_model = SampleNet(TwoFC())
        client_models = self._create_n_clients(num_total_users)
        torch.manual_seed(1)
        np.random.seed(1)
        global_model_trained1 = self.train_async_with_zero_weight(
            initial_model=copy.deepcopy(initial_model),
            client_models=client_models,
            num_epochs=num_epochs,
            num_total_users=num_total_users,
            momentum=momentum,
            train_with_zero_weight_in_middle=False,
        )
        torch.manual_seed(1)
        np.random.seed(1)
        global_model_trained2 = self.train_async_with_zero_weight(
            initial_model=copy.deepcopy(initial_model),
            client_models=client_models,
            num_epochs=num_epochs,
            num_total_users=num_total_users,
            momentum=momentum,
            train_with_zero_weight_in_middle=True,
        )

        error_msg = verify_models_equivalent_after_training(
            global_model_trained1,
            global_model_trained2,
            initial_model,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEqual(error_msg, "")

    @pytest.mark.parametrize(
        "num_total_users,num_epochs, momentum, lr",
        [(1, 2, 0.5, 10), (10, 10, 0.5, 10), (10, 10, 0, 10)],
    )
    def test_momentum_implementation_one_weight(
        self, num_total_users, num_epochs, momentum, lr
    ):
        """FedAsyncAggregatorWithMomentum.on_client_training_end should behave
        exactly like SGD with momentum when weight = 1
        We test this by
        a) Running SGD with momentum
        b) Running AsyncFL in sequential model with momentum
        Showing that (a) and (b) produce the same results
        """
        momentum = 0.5
        num_epochs = 10
        num_total_users = 10
        lr = 1.0

        initial_model = SampleNet(TwoFC())
        client_models = self._create_n_clients(num_total_users)

        # run async training
        torch.manual_seed(1)
        np.random.seed(1)
        config = FedAvgWithLRWithMomentumAsyncAggregatorConfig(lr=lr, momentum=momentum)
        aggregator = instantiate(config, global_model=copy.deepcopy(initial_model))
        for _ in range(num_epochs):
            for client in client_models:
                aggregator.on_client_training_end(
                    client.delta, client.after_train, weight=1
                )

        # run SGD training
        torch.manual_seed(1)
        np.random.seed(1)
        sgd_model = copy.deepcopy(initial_model)
        sgd_optimizer = torch.optim.SGD(
            sgd_model.fl_get_module().parameters(), lr=lr, momentum=momentum
        )

        for _ in range(num_epochs):
            for client in client_models:
                FLModelParamUtils.set_gradient(
                    model=sgd_model.fl_get_module(),
                    reference_gradient=client.delta.fl_get_module(),
                )
                sgd_optimizer.step()

        error_msg = verify_models_equivalent_after_training(
            aggregator.global_model,
            sgd_model,
            initial_model,
            rel_epsilon=1e-6,
            abs_epsilon=1e-6,
        )
        assertEqual(error_msg, "")
